from typing import Optional, Sequence

from pydantic import BaseModel, ConfigDict, Field


class JobSummary(BaseModel):
    """Summary of job execution counts for a SQL query."""

    success_job_ids: Sequence[int] = Field(..., alias="successJobsIds")
    failed_job_ids: Sequence[int] = Field(..., alias="failedJobsIds")
    running_job_ids: Sequence[int] = Field(..., alias="runningJobsIds")

    model_config = ConfigDict(populate_by_name=True)


class SqlQuerySummary(BaseModel):
    """Simplified summary of a SQL query execution for LLM consumption."""

    id: int
    duration: Optional[int] = None  # Duration in milliseconds
    description: Optional[str] = None
    status: str
    submission_time: Optional[str] = Field(None, alias="submissionTime")
    plan_description: str = Field(..., alias="planDescription")
    job_summary: JobSummary = Field(..., alias="jobSummary")

    model_config = ConfigDict(populate_by_name=True)



from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Sequence, Set

from pydantic import BaseModel, ConfigDict, Field, field_validator


class JobExecutionStatus(str, Enum):
    RUNNING = "RUNNING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_string(cls, s: str) -> "JobExecutionStatus":
        try:
            return cls(s.upper())
        except ValueError:
            return cls[s.upper()]


class StageStatus(str, Enum):
    PENDING = "PENDING"
    ACTIVE = "ACTIVE"
    COMPLETE = "COMPLETE"
    SKIPPED = "SKIPPED"
    FAILED = "FAILED"

    @classmethod
    def from_string(cls, s: str) -> "StageStatus":
        try:
            return cls(s.upper())
        except ValueError:
            return cls[s.upper()]


class TaskStatus(str, Enum):
    RUNNING = "RUNNING"
    KILLED = "KILLED"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_string(cls, s: str) -> "TaskStatus":
        try:
            return cls(s.upper())
        except ValueError:
            return cls[s.upper()]


class TaskSorting(str, Enum):
    ID = "ID"
    INCREASING_RUNTIME = "INCREASING_RUNTIME"
    DECREASING_RUNTIME = "DECREASING_RUNTIME"

    _alternate_names = {"runtime": INCREASING_RUNTIME, "-runtime": DECREASING_RUNTIME}

    @classmethod
    def from_string(cls, s: str) -> "TaskSorting":
        lower = s.lower()
        if lower in cls._alternate_names:
            return cls._alternate_names[lower]
        try:
            return cls(s.upper())
        except ValueError:
            return cls[s.upper()]


class ApplicationStatus(str, Enum):
    COMPLETED = "COMPLETED"
    RUNNING = "RUNNING"

    @classmethod
    def from_string(cls, s: str) -> "ApplicationStatus":
        try:
            return cls(s.upper())
        except ValueError:
            return cls[s.upper()]


class ThreadState(str, Enum):
    NEW = "NEW"
    RUNNABLE = "RUNNABLE"
    BLOCKED = "BLOCKED"
    WAITING = "WAITING"
    TIMED_WAITING = "TIMED_WAITING"
    TERMINATED = "TERMINATED"


class ExecutorMetrics(BaseModel):
    metrics: Optional[Dict[str, int]] = Field(None, alias="metrics")

    model_config = ConfigDict(populate_by_name=True)


class ApplicationInfo(BaseModel):
    id: str
    name: str
    cores_granted: Optional[int] = Field(None, alias="coresGranted")
    max_cores: Optional[int] = Field(None, alias="maxCores")
    cores_per_executor: Optional[int] = Field(None, alias="coresPerExecutor")
    memory_per_executor_mb: Optional[int] = Field(None, alias="memoryPerExecutorMB")
    attempts: Sequence["ApplicationAttemptInfo"]

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class ApplicationAttemptInfo(BaseModel):
    attempt_id: Optional[str] = Field(None, alias="attemptId")
    start_time: Optional[datetime] = Field(None, alias="startTime")
    end_time: Optional[datetime] = Field(None, alias="endTime")
    last_updated: Optional[datetime] = Field(None, alias="lastUpdated")
    duration: int
    spark_user: Optional[str] = Field(None, alias="sparkUser")
    app_spark_version: Optional[str] = Field(None, alias="appSparkVersion")
    completed: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @field_validator("start_time", "end_time", "last_updated", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value


class ResourceProfileInfo(BaseModel):
    id: int
    executor_resources: Optional[Dict[str, Any]] = Field(
        None, alias="executorResources"
    )  # Will be typed properly once those classes are defined
    task_resources: Optional[Dict[str, Any]] = Field(None, alias="taskResources")

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class ExecutorStageSummary(BaseModel):
    task_time: Optional[int] = Field(None, alias="taskTime")
    failed_tasks: Optional[int] = Field(None, alias="failedTasks")
    succeeded_tasks: Optional[int] = Field(None, alias="succeededTasks")
    killed_tasks: Optional[int] = Field(None, alias="killedTasks")
    input_bytes: Optional[int] = Field(None, alias="inputBytes")
    input_records: Optional[int] = Field(None, alias="inputRecords")
    output_bytes: Optional[int] = Field(None, alias="outputBytes")
    output_records: Optional[int] = Field(None, alias="outputRecords")
    shuffle_read: Optional[int] = Field(None, alias="shuffleRead")
    shuffle_read_records: Optional[int] = Field(None, alias="shuffleReadRecords")
    shuffle_write: Optional[int] = Field(None, alias="shuffleWrite")
    shuffle_write_records: Optional[int] = Field(None, alias="shuffleWriteRecords")
    memory_bytes_spilled: Optional[int] = Field(None, alias="memoryBytesSpilled")
    disk_bytes_spilled: Optional[int] = Field(None, alias="diskBytesSpilled")
    is_blacklisted_for_stage: Optional[bool] = Field(
        None, alias="isBlacklistedForStage"
    )  # deprecated
    peak_memory_metrics: Optional[ExecutorMetrics] = Field(
        None, alias="peakMemoryMetrics"
    )
    is_excluded_for_stage: Optional[bool] = Field(None, alias="isExcludedForStage")

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class SpeculationStageSummary(BaseModel):
    num_tasks: Optional[int] = Field(None, alias="numTasks")
    num_active_tasks: Optional[int] = Field(None, alias="numActiveTasks")
    num_completed_tasks: Optional[int] = Field(None, alias="numCompletedTasks")
    num_failed_tasks: Optional[int] = Field(None, alias="numFailedTasks")
    num_killed_tasks: Optional[int] = Field(None, alias="numKilledTasks")

    model_config = ConfigDict(populate_by_name=True)


class ExecutorSummary(BaseModel):
    id: str
    host_port: Optional[str] = Field(None, alias="hostPort")
    is_active: Optional[bool] = Field(None, alias="isActive")
    rdd_blocks: Optional[int] = Field(None, alias="rddBlocks")
    memory_used: Optional[int] = Field(None, alias="memoryUsed")
    disk_used: Optional[int] = Field(None, alias="diskUsed")
    total_cores: Optional[int] = Field(None, alias="totalCores")
    max_tasks: Optional[int] = Field(None, alias="maxTasks")
    active_tasks: Optional[int] = Field(None, alias="activeTasks")
    failed_tasks: Optional[int] = Field(None, alias="failedTasks")
    completed_tasks: Optional[int] = Field(None, alias="completedTasks")
    total_tasks: Optional[int] = Field(None, alias="totalTasks")
    total_duration: Optional[int] = Field(None, alias="totalDuration")
    total_gc_time: Optional[int] = Field(None, alias="totalGCTime")
    total_input_bytes: Optional[int] = Field(None, alias="totalInputBytes")
    total_shuffle_read: Optional[int] = Field(None, alias="totalShuffleRead")
    total_shuffle_write: Optional[int] = Field(None, alias="totalShuffleWrite")
    is_blacklisted: Optional[bool] = Field(None, alias="isBlacklisted")  # deprecated
    max_memory: Optional[int] = Field(None, alias="maxMemory")
    add_time: Optional[datetime] = Field(None, alias="addTime")
    remove_time: Optional[datetime] = Field(None, alias="removeTime")
    remove_reason: Optional[str] = Field(None, alias="removeReason")
    executor_logs: Optional[Dict[str, str]] = Field(None, alias="executorLogs")
    memory_metrics: Optional["MemoryMetrics"] = Field(None, alias="memoryMetrics")
    blacklisted_in_stages: Set[int] = Field(
        set(), alias="blacklistedInStages"
    )  # deprecated
    peak_memory_metrics: Optional[ExecutorMetrics] = Field(
        None, alias="peakMemoryMetrics"
    )
    attributes: Dict[str, str]
    resources: Dict[
        str, Any
    ]  # Will be typed properly once ResourceInformation is defined
    resource_profile_id: Optional[int] = Field(None, alias="resourceProfileId")
    is_excluded: Optional[bool] = Field(None, alias="isExcluded")
    excluded_in_stages: Set[int] = Field(set(), alias="excludedInStages")

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @field_validator("add_time", "remove_time", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if value is None:
            return None
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value


class MemoryMetrics(BaseModel):
    used_on_heap_storage_memory: Optional[int] = Field(
        None, alias="usedOnHeapStorageMemory"
    )
    used_off_heap_storage_memory: Optional[int] = Field(
        None, alias="usedOffHeapStorageMemory"
    )
    total_on_heap_storage_memory: Optional[int] = Field(
        None, alias="totalOnHeapStorageMemory"
    )
    total_off_heap_storage_memory: Optional[int] = Field(
        None, alias="totalOffHeapStorageMemory"
    )

    model_config = ConfigDict(populate_by_name=True)


class JobData(BaseModel):
    job_id: Optional[int] = Field(None, alias="jobId")
    name: str
    description: Optional[str] = None
    submission_time: Optional[datetime] = Field(None, alias="submissionTime")
    completion_time: Optional[datetime] = Field(None, alias="completionTime")
    stage_ids: Optional[Sequence[int]] = Field(None, alias="stageIds")
    job_group: Optional[str] = Field(None, alias="jobGroup")
    job_tags: Sequence[str] = Field([], alias="jobTags")
    status: str  # JobExecutionStatus as string
    num_tasks: Optional[int] = Field(None, alias="numTasks")
    num_active_tasks: Optional[int] = Field(None, alias="numActiveTasks")
    num_completed_tasks: Optional[int] = Field(None, alias="numCompletedTasks")
    num_skipped_tasks: Optional[int] = Field(None, alias="numSkippedTasks")
    num_failed_tasks: Optional[int] = Field(None, alias="numFailedTasks")
    num_killed_tasks: Optional[int] = Field(None, alias="numKilledTasks")
    num_completed_indices: Optional[int] = Field(None, alias="numCompletedIndices")
    num_active_stages: Optional[int] = Field(None, alias="numActiveStages")
    num_completed_stages: Optional[int] = Field(None, alias="numCompletedStages")
    num_skipped_stages: Optional[int] = Field(None, alias="numSkippedStages")
    num_failed_stages: Optional[int] = Field(None, alias="numFailedStages")
    killed_tasks_summary: Dict[str, int] = Field({}, alias="killedTasksSummary")

    model_config = ConfigDict(
        populate_by_name=True, arbitrary_types_allowed=True, extra="ignore"
    )

    @field_validator("submission_time", "completion_time", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if value is None:
            return None
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value


class RDDStorageInfo(BaseModel):
    id: int
    name: str
    num_partitions: Optional[int] = Field(None, alias="numPartitions")
    num_cached_partitions: Optional[int] = Field(None, alias="numCachedPartitions")
    storage_level: Optional[str] = Field(None, alias="storageLevel")
    memory_used: Optional[int] = Field(None, alias="memoryUsed")
    disk_used: Optional[int] = Field(None, alias="diskUsed")
    data_distribution: Optional[Sequence["RDDDataDistribution"]] = Field(
        None, alias="dataDistribution"
    )
    partitions: Optional[Sequence["RDDPartitionInfo"]] = None

    model_config = ConfigDict(
        populate_by_name=True,
        arbitrary_types_allowed=True,
        json_encoders={datetime: lambda v: v.isoformat() if v else None},
    )


class RDDDataDistribution(BaseModel):
    address: str
    memory_used: Optional[int] = Field(None, alias="memoryUsed")
    memory_remaining: Optional[int] = Field(None, alias="memoryRemaining")
    disk_used: Optional[int] = Field(None, alias="diskUsed")
    on_heap_memory_used: Optional[int] = Field(None, alias="onHeapMemoryUsed")
    off_heap_memory_used: Optional[int] = Field(None, alias="offHeapMemoryUsed")
    on_heap_memory_remaining: Optional[int] = Field(None, alias="onHeapMemoryRemaining")
    off_heap_memory_remaining: Optional[int] = Field(
        None, alias="offHeapMemoryRemaining"
    )

    model_config = ConfigDict(populate_by_name=True)


class RDDPartitionInfo(BaseModel):
    block_name: Optional[str] = Field(None, alias="blockName")
    storage_level: Optional[str] = Field(None, alias="storageLevel")
    memory_used: Optional[int] = Field(None, alias="memoryUsed")
    disk_used: Optional[int] = Field(None, alias="diskUsed")
    executors: Sequence[str]

    model_config = ConfigDict(populate_by_name=True)


class StageData(BaseModel):
    status: str  # StageStatus as string
    stage_id: Optional[int] = Field(None, alias="stageId")
    attempt_id: Optional[int] = Field(None, alias="attemptId")
    num_tasks: Optional[int] = Field(None, alias="numTasks")
    num_active_tasks: Optional[int] = Field(None, alias="numActiveTasks")
    num_complete_tasks: Optional[int] = Field(None, alias="numCompleteTasks")
    num_failed_tasks: Optional[int] = Field(None, alias="numFailedTasks")
    num_killed_tasks: Optional[int] = Field(None, alias="numKilledTasks")
    num_completed_indices: Optional[int] = Field(None, alias="numCompletedIndices")

    submission_time: Optional[datetime] = Field(None, alias="submissionTime")
    first_task_launched_time: Optional[datetime] = Field(
        None, alias="firstTaskLaunchedTime"
    )
    completion_time: Optional[datetime] = Field(None, alias="completionTime")
    failure_reason: Optional[str] = Field(None, alias="failureReason")

    executor_deserialize_time: Optional[int] = Field(
        None, alias="executorDeserializeTime"
    )
    executor_deserialize_cpu_time: Optional[int] = Field(
        None, alias="executorDeserializeCpuTime"
    )
    executor_run_time: Optional[int] = Field(None, alias="executorRunTime")
    executor_cpu_time: Optional[int] = Field(None, alias="executorCpuTime")
    result_size: Optional[int] = Field(None, alias="resultSize")
    jvm_gc_time: Optional[int] = Field(None, alias="jvmGcTime")
    result_serialization_time: Optional[int] = Field(
        None, alias="resultSerializationTime"
    )
    memory_bytes_spilled: Optional[int] = Field(None, alias="memoryBytesSpilled")
    disk_bytes_spilled: Optional[int] = Field(None, alias="diskBytesSpilled")
    peak_execution_memory: Optional[int] = Field(None, alias="peakExecutionMemory")
    input_bytes: Optional[int] = Field(None, alias="inputBytes")
    input_records: Optional[int] = Field(None, alias="inputRecords")
    output_bytes: Optional[int] = Field(None, alias="outputBytes")
    output_records: Optional[int] = Field(None, alias="outputRecords")
    shuffle_remote_blocks_fetched: Optional[int] = Field(
        None, alias="shuffleRemoteBlocksFetched"
    )
    shuffle_local_blocks_fetched: Optional[int] = Field(
        None, alias="shuffleLocalBlocksFetched"
    )
    shuffle_fetch_wait_time: Optional[int] = Field(None, alias="shuffleFetchWaitTime")
    shuffle_remote_bytes_read: Optional[int] = Field(
        None, alias="shuffleRemoteBytesRead"
    )
    shuffle_remote_bytes_read_to_disk: Optional[int] = Field(
        None, alias="shuffleRemoteBytesReadToDisk"
    )
    shuffle_local_bytes_read: Optional[int] = Field(None, alias="shuffleLocalBytesRead")
    shuffle_read_bytes: Optional[int] = Field(None, alias="shuffleReadBytes")
    shuffle_read_records: Optional[int] = Field(None, alias="shuffleReadRecords")
    shuffle_corrupt_merged_block_chunks: Optional[int] = Field(
        0, alias="shuffleCorruptMergedBlockChunks"
    )
    shuffle_merged_fetch_fallback_count: Optional[int] = Field(
        0, alias="shuffleMergedFetchFallbackCount"
    )
    shuffle_merged_remote_blocks_fetched: Optional[int] = Field(
        0, alias="shuffleMergedRemoteBlocksFetched"
    )
    shuffle_merged_local_blocks_fetched: Optional[int] = Field(
        0, alias="shuffleMergedLocalBlocksFetched"
    )
    shuffle_merged_remote_chunks_fetched: Optional[int] = Field(
        0, alias="shuffleMergedRemoteChunksFetched"
    )
    shuffle_merged_local_chunks_fetched: Optional[int] = Field(
        0, alias="shuffleMergedLocalChunksFetched"
    )
    shuffle_merged_remote_bytes_read: Optional[int] = Field(
        0, alias="shuffleMergedRemoteBytesRead"
    )
    shuffle_merged_local_bytes_read: Optional[int] = Field(
        0, alias="shuffleMergedLocalBytesRead"
    )
    shuffle_remote_reqs_duration: Optional[int] = Field(
        0, alias="shuffleRemoteReqsDuration"
    )
    shuffle_merged_remote_reqs_duration: Optional[int] = Field(
        0, alias="shuffleMergedRemoteReqsDuration"
    )
    shuffle_write_bytes: Optional[int] = Field(None, alias="shuffleWriteBytes")
    shuffle_write_time: Optional[int] = Field(None, alias="shuffleWriteTime")
    shuffle_write_records: Optional[int] = Field(None, alias="shuffleWriteRecords")

    name: str
    description: Optional[str] = None
    details: str
    scheduling_pool: Optional[str] = Field(None, alias="schedulingPool")

    # rdd_ids: Optional[Sequence[int]] = Field(None, alias="rddIds")
    accumulator_updates: Optional[Sequence["AccumulableInfo"]] = Field(
        None, alias="accumulatorUpdates"
    )
    tasks: Optional[Dict[str, "TaskData"]] = None
    executor_summary: Optional[Dict[str, ExecutorStageSummary]] = Field(
        None, alias="executorSummary"
    )
    speculation_summary: Optional[SpeculationStageSummary] = Field(
        None, alias="speculationSummary"
    )
    killed_tasks_summary: Dict[str, int] = Field({}, alias="killedTasksSummary")
    resource_profile_id: Optional[int] = Field(None, alias="resourceProfileId")
    peak_executor_metrics: Optional[ExecutorMetrics] = Field(
        None, alias="peakExecutorMetrics"
    )
    task_metrics_distributions: Optional["TaskMetricDistributions"] = Field(
        None, alias="taskMetricsDistributions"
    )
    executor_metrics_distributions: Optional["ExecutorMetricsDistributions"] = Field(
        None, alias="executorMetricsDistributions"
    )
    is_shuffle_push_enabled: bool = Field(False, alias="isShufflePushEnabled")
    shuffle_mergers_count: Optional[int] = Field(0, alias="shuffleMergersCount")

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)

    @field_validator(
        "submission_time", "first_task_launched_time", "completion_time", mode="before"
    )
    @classmethod
    def parse_datetime(cls, value):
        if value is None:
            return None
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value


class TaskData(BaseModel):
    task_id: Optional[int] = Field(None, alias="taskId")
    index: int
    attempt: int
    partition_id: Optional[int] = Field(None, alias="partitionId")
    launch_time: Optional[datetime] = Field(None, alias="launchTime")
    result_fetch_start: Optional[datetime] = Field(None, alias="resultFetchStart")
    duration: Optional[int] = None
    executor_id: Optional[str] = Field(None, alias="executorId")
    host: str
    status: str
    task_locality: Optional[str] = Field(None, alias="taskLocality")
    speculative: bool
    accumulator_updates: Optional[Sequence["AccumulableInfo"]] = Field(
        None, alias="accumulatorUpdates"
    )
    error_message: Optional[str] = Field(None, alias="errorMessage")
    task_metrics: Optional["TaskMetrics"] = Field(None, alias="taskMetrics")
    executor_logs: Dict[str, str] = Field({}, alias="executorLogs")
    scheduler_delay: Optional[int] = Field(0, alias="schedulerDelay")
    getting_result_time: Optional[int] = Field(0, alias="gettingResultTime")

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)

    @field_validator("launch_time", "result_fetch_start", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if value is None:
            return None
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value


class TaskMetrics(BaseModel):
    executor_deserialize_time: Optional[int] = Field(
        None, alias="executorDeserializeTime"
    )
    executor_deserialize_cpu_time: Optional[int] = Field(
        None, alias="executorDeserializeCpuTime"
    )
    executor_run_time: Optional[int] = Field(None, alias="executorRunTime")
    executor_cpu_time: Optional[int] = Field(None, alias="executorCpuTime")
    result_size: Optional[int] = Field(None, alias="resultSize")
    jvm_gc_time: Optional[int] = Field(None, alias="jvmGcTime")
    result_serialization_time: Optional[int] = Field(
        None, alias="resultSerializationTime"
    )
    memory_bytes_spilled: Optional[int] = Field(None, alias="memoryBytesSpilled")
    disk_bytes_spilled: Optional[int] = Field(None, alias="diskBytesSpilled")
    peak_execution_memory: Optional[int] = Field(None, alias="peakExecutionMemory")
    input_metrics: Optional["InputMetrics"] = Field(None, alias="inputMetrics")
    output_metrics: Optional["OutputMetrics"] = Field(None, alias="outputMetrics")
    shuffle_read_metrics: Optional["ShuffleReadMetrics"] = Field(
        None, alias="shuffleReadMetrics"
    )
    shuffle_write_metrics: Optional["ShuffleWriteMetrics"] = Field(
        None, alias="shuffleWriteMetrics"
    )

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)


class InputMetrics(BaseModel):
    bytes_read: Optional[int] = Field(None, alias="bytesRead")
    records_read: Optional[int] = Field(None, alias="recordsRead")

    model_config = ConfigDict(populate_by_name=True)


class OutputMetrics(BaseModel):
    bytes_written: Optional[int] = Field(None, alias="bytesWritten")
    records_written: Optional[int] = Field(None, alias="recordsWritten")

    model_config = ConfigDict(populate_by_name=True)


class ShufflePushReadMetrics(BaseModel):
    corrupt_merged_block_chunks: Optional[int] = Field(
        None, alias="corruptMergedBlockChunks"
    )
    merged_fetch_fallback_count: Optional[int] = Field(
        None, alias="mergedFetchFallbackCount"
    )
    remote_merged_blocks_fetched: Optional[int] = Field(
        None, alias="remoteMergedBlocksFetched"
    )
    local_merged_blocks_fetched: Optional[int] = Field(
        None, alias="localMergedBlocksFetched"
    )
    remote_merged_chunks_fetched: Optional[int] = Field(
        None, alias="remoteMergedChunksFetched"
    )
    local_merged_chunks_fetched: Optional[int] = Field(
        None, alias="localMergedChunksFetched"
    )
    remote_merged_bytes_read: Optional[int] = Field(None, alias="remoteMergedBytesRead")
    local_merged_bytes_read: Optional[int] = Field(None, alias="localMergedBytesRead")
    remote_merged_reqs_duration: Optional[int] = Field(
        None, alias="remoteMergedReqsDuration"
    )

    model_config = ConfigDict(populate_by_name=True)


class ShuffleReadMetrics(BaseModel):
    remote_blocks_fetched: Optional[int] = Field(None, alias="remoteBlocksFetched")
    local_blocks_fetched: Optional[int] = Field(None, alias="localBlocksFetched")
    fetch_wait_time: Optional[int] = Field(None, alias="fetchWaitTime")
    remote_bytes_read: Optional[int] = Field(None, alias="remoteBytesRead")
    remote_bytes_read_to_disk: Optional[int] = Field(
        None, alias="remoteBytesReadToDisk"
    )
    local_bytes_read: Optional[int] = Field(None, alias="localBytesRead")
    records_read: Optional[int] = Field(None, alias="recordsRead")
    remote_reqs_duration: Optional[int] = Field(None, alias="remoteReqsDuration")
    shuffle_push_read_metrics: Optional[ShufflePushReadMetrics] = Field(
        None, alias="shufflePushReadMetrics"
    )

    model_config = ConfigDict(populate_by_name=True)


class ShuffleWriteMetrics(BaseModel):
    bytes_written: Optional[int] = Field(None, alias="bytesWritten")
    write_time: Optional[int] = Field(None, alias="writeTime")
    records_written: Optional[int] = Field(None, alias="recordsWritten")

    model_config = ConfigDict(populate_by_name=True)


class TaskMetricDistributions(BaseModel):
    quantiles: Optional[Sequence[float]] = Field(None, alias="quantiles")

    duration: Optional[Sequence[float]] = Field(None, alias="duration")
    executor_deserialize_time: Optional[Sequence[float]] = Field(
        None, alias="executorDeserializeTime"
    )
    executor_deserialize_cpu_time: Optional[Sequence[float]] = Field(
        None, alias="executorDeserializeCpuTime"
    )
    executor_run_time: Optional[Sequence[float]] = Field(None, alias="executorRunTime")
    executor_cpu_time: Optional[Sequence[float]] = Field(None, alias="executorCpuTime")
    result_size: Optional[Sequence[float]] = Field(None, alias="resultSize")
    jvm_gc_time: Optional[Sequence[float]] = Field(None, alias="jvmGcTime")
    result_serialization_time: Optional[Sequence[float]] = Field(
        None, alias="resultSerializationTime"
    )
    getting_result_time: Optional[Sequence[float]] = Field(
        None, alias="gettingResultTime"
    )
    scheduler_delay: Optional[Sequence[float]] = Field(None, alias="schedulerDelay")
    peak_execution_memory: Optional[Sequence[float]] = Field(
        None, alias="peakExecutionMemory"
    )
    memory_bytes_spilled: Optional[Sequence[float]] = Field(
        None, alias="memoryBytesSpilled"
    )
    disk_bytes_spilled: Optional[Sequence[float]] = Field(
        None, alias="diskBytesSpilled"
    )

    input_metrics: Optional["InputMetricDistributions"] = Field(
        None, alias="inputMetrics"
    )
    output_metrics: Optional["OutputMetricDistributions"] = Field(
        None, alias="outputMetrics"
    )
    shuffle_read_metrics: Optional["ShuffleReadMetricDistributions"] = Field(
        None, alias="shuffleReadMetrics"
    )
    shuffle_write_metrics: Optional["ShuffleWriteMetricDistributions"] = Field(
        None, alias="shuffleWriteMetrics"
    )

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)


class InputMetricDistributions(BaseModel):
    bytes_read: Optional[Sequence[float]] = Field(None, alias="bytesRead")
    records_read: Optional[Sequence[float]] = Field(None, alias="recordsRead")

    model_config = ConfigDict(populate_by_name=True)


class OutputMetricDistributions(BaseModel):
    bytes_written: Optional[Sequence[float]] = Field(None, alias="bytesWritten")
    records_written: Optional[Sequence[float]] = Field(None, alias="recordsWritten")

    model_config = ConfigDict(populate_by_name=True)


class ShufflePushReadMetricDistributions(BaseModel):
    corrupt_merged_block_chunks: Optional[Sequence[float]] = Field(
        None, alias="corruptMergedBlockChunks"
    )
    merged_fetch_fallback_count: Optional[Sequence[float]] = Field(
        None, alias="mergedFetchFallbackCount"
    )
    remote_merged_blocks_fetched: Optional[Sequence[float]] = Field(
        None, alias="remoteMergedBlocksFetched"
    )
    local_merged_blocks_fetched: Optional[Sequence[float]] = Field(
        None, alias="localMergedBlocksFetched"
    )
    remote_merged_chunks_fetched: Optional[Sequence[float]] = Field(
        None, alias="remoteMergedChunksFetched"
    )
    local_merged_chunks_fetched: Optional[Sequence[float]] = Field(
        None, alias="localMergedChunksFetched"
    )
    remote_merged_bytes_read: Optional[Sequence[float]] = Field(
        None, alias="remoteMergedBytesRead"
    )
    local_merged_bytes_read: Optional[Sequence[float]] = Field(
        None, alias="localMergedBytesRead"
    )
    remote_merged_reqs_duration: Optional[Sequence[float]] = Field(
        None, alias="remoteMergedReqsDuration"
    )

    model_config = ConfigDict(populate_by_name=True)


class ExecutorMetricsDistributions(BaseModel):
    quantiles: Sequence[float]

    task_time: Optional[Sequence[float]] = Field(None, alias="taskTime")
    failed_tasks: Optional[Sequence[float]] = Field(None, alias="failedTasks")
    succeeded_tasks: Optional[Sequence[float]] = Field(None, alias="succeededTasks")
    killed_tasks: Optional[Sequence[float]] = Field(None, alias="killedTasks")
    input_bytes: Optional[Sequence[float]] = Field(None, alias="inputBytes")
    input_records: Optional[Sequence[float]] = Field(None, alias="inputRecords")
    output_bytes: Optional[Sequence[float]] = Field(None, alias="outputBytes")
    output_records: Optional[Sequence[float]] = Field(None, alias="outputRecords")
    shuffle_read: Optional[Sequence[float]] = Field(None, alias="shuffleRead")
    shuffle_read_records: Optional[Sequence[float]] = Field(
        None, alias="shuffleReadRecords"
    )
    shuffle_write: Optional[Sequence[float]] = Field(None, alias="shuffleWrite")
    shuffle_write_records: Optional[Sequence[float]] = Field(
        None, alias="shuffleWriteRecords"
    )
    memory_bytes_spilled: Optional[Sequence[float]] = Field(
        None, alias="memoryBytesSpilled"
    )
    disk_bytes_spilled: Optional[Sequence[float]] = Field(
        None, alias="diskBytesSpilled"
    )
    peak_memory_metrics: Optional["ExecutorPeakMetricsDistributions"] = Field(
        None, alias="peakMemoryMetrics"
    )

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)


class ExecutorPeakMetricsDistributions(BaseModel):
    quantiles: Optional[Sequence[float]] = None
    executor_metrics: Optional[Sequence[ExecutorMetrics]] = Field(
        None, alias="executorMetrics"
    )

    model_config = ConfigDict(populate_by_name=True)


class ShuffleReadMetricDistributions(BaseModel):
    read_bytes: Optional[Sequence[float]] = Field(None, alias="readBytes")
    read_records: Optional[Sequence[float]] = Field(None, alias="readRecords")
    remote_blocks_fetched: Optional[Sequence[float]] = Field(
        None, alias="remoteBlocksFetched"
    )
    local_blocks_fetched: Optional[Sequence[float]] = Field(
        None, alias="localBlocksFetched"
    )
    fetch_wait_time: Optional[Sequence[float]] = Field(None, alias="fetchWaitTime")
    remote_bytes_read: Optional[Sequence[float]] = Field(None, alias="remoteBytesRead")
    remote_bytes_read_to_disk: Optional[Sequence[float]] = Field(
        None, alias="remoteBytesReadToDisk"
    )
    total_blocks_fetched: Optional[Sequence[float]] = Field(
        None, alias="totalBlocksFetched"
    )
    remote_reqs_duration: Optional[Sequence[float]] = Field(
        None, alias="remoteReqsDuration"
    )
    shuffle_push_read_metrics_dist: Optional[ShufflePushReadMetricDistributions] = (
        Field(None, alias="shufflePushReadMetricsDist")
    )

    model_config = ConfigDict(populate_by_name=True)


class ShuffleWriteMetricDistributions(BaseModel):
    write_bytes: Optional[Sequence[float]] = Field(None, alias="writeBytes")
    write_records: Optional[Sequence[float]] = Field(None, alias="writeRecords")
    write_time: Optional[Sequence[float]] = Field(None, alias="writeTime")

    model_config = ConfigDict(populate_by_name=True)


class AccumulableInfo(BaseModel):
    id: int
    name: str
    update: str = None
    value: str

    model_config = ConfigDict(populate_by_name=True)


class VersionInfo(BaseModel):
    spark: str

    model_config = ConfigDict(populate_by_name=True)


class ApplicationEnvironmentInfo(BaseModel):
    runtime: "RuntimeInfo"
    spark_properties: Optional[Sequence[tuple[str, str]]] = Field(
        None, alias="sparkProperties"
    )
    hadoop_properties: Optional[Sequence[tuple[str, str]]] = Field(
        None, alias="hadoopProperties"
    )
    system_properties: Optional[Sequence[tuple[str, str]]] = Field(
        None, alias="systemProperties"
    )
    metrics_properties: Optional[Sequence[tuple[str, str]]] = Field(
        None, alias="metricsProperties"
    )
    classpath_entries: Optional[Sequence[tuple[str, str]]] = Field(
        None, alias="classpathEntries"
    )
    resource_profiles: Optional[Sequence[ResourceProfileInfo]] = Field(
        None, alias="resourceProfiles"
    )

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)


class RuntimeInfo(BaseModel):
    java_version: Optional[str] = Field(None, alias="javaVersion")
    java_home: Optional[str] = Field(None, alias="javaHome")
    scala_version: Optional[str] = Field(None, alias="scalaVersion")

    model_config = ConfigDict(populate_by_name=True)


class StackTrace(BaseModel):
    elems: Sequence[str]

    def __str__(self) -> str:
        return "".join(self.elems)

    def html(self) -> str:
        return "<br />".join(elem.rstrip() for elem in self.elems)

    def mkstring(self, start: str, sep: str, end: str) -> str:
        return start + sep.join(self.elems) + end

    model_config = ConfigDict(populate_by_name=True)


class ThreadStackTrace(BaseModel):
    thread_id: Optional[int] = Field(None, alias="threadId")
    thread_name: Optional[str] = Field(None, alias="threadName")
    thread_state: Optional[str] = Field(
        None, alias="threadState"
    )  # ThreadState as string
    stack_trace: Optional[StackTrace] = Field(None, alias="stackTrace")
    blocked_by_thread_id: Optional[int] = Field(None, alias="blockedByThreadId")
    blocked_by_lock: Optional[str] = Field(None, alias="blockedByLock")
    holding_locks: Sequence[str] = Field([], alias="holdingLocks")  # deprecated
    synchronizers: Sequence[str]
    monitors: Sequence[str]
    lock_name: Optional[str] = Field(None, alias="lockName")
    lock_owner_name: Optional[str] = Field(None, alias="lockOwnerName")
    suspended: bool
    in_native: Optional[bool] = Field(None, alias="inNative")
    is_daemon: Optional[bool] = Field(None, alias="isDaemon")
    priority: int

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)


class ProcessSummary(BaseModel):
    id: str
    host_port: Optional[str] = Field(None, alias="hostPort")
    is_active: Optional[bool] = Field(None, alias="isActive")
    total_cores: Optional[int] = Field(None, alias="totalCores")
    add_time: Optional[datetime] = Field(None, alias="addTime")
    remove_time: Optional[datetime] = Field(None, alias="removeTime")
    process_logs: Optional[Dict[str, str]] = Field(None, alias="processLogs")

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)

    @field_validator("add_time", "remove_time", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if value is None:
            return None
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value


class SQLExecutionStatus(str, Enum):
    """Represents the status of a SQL execution."""

    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class Metric(BaseModel):
    """Represents a metric in a SQL execution plan node."""

    name: str
    value: str

    model_config = ConfigDict(populate_by_name=True)


class Node(BaseModel):
    """Represents a node in a SQL execution plan."""

    node_id: int = Field(..., alias="nodeId")
    node_name: str = Field(..., alias="nodeName")
    whole_stage_codegen_id: Optional[int] = Field(None, alias="wholeStageCodegenId")
    metrics: Sequence[Metric]

    model_config = ConfigDict(populate_by_name=True)


class SparkPlanGraphEdge(BaseModel):
    """Represents an edge in a SQL execution plan graph."""

    from_id: int = Field(..., alias="fromId")
    to_id: int = Field(..., alias="toId")

    model_config = ConfigDict(populate_by_name=True)


class ExecutionData(BaseModel):
    """Represents data about a SQL execution."""

    id: int
    status: str  # SQLExecutionStatus as string
    description: Optional[str] = Field(None, alias="description")
    plan_description: str = Field(..., alias="planDescription")
    submission_time: datetime = Field(..., alias="submissionTime")
    duration: Optional[int] = Field(None, alias="durationMilliSeconds")
    running_job_ids: Sequence[int] = Field([], alias="runningJobIds")
    success_job_ids: Sequence[int] = Field([], alias="successJobIds")
    failed_job_ids: Sequence[int] = Field([], alias="failedJobIds")
    nodes: Sequence[Node]
    edges: Sequence[SparkPlanGraphEdge]

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)

    @field_validator("submission_time", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if isinstance(value, (int, float)):
            return datetime.fromtimestamp(value / 1000)
        if isinstance(value, str) and value.endswith("GMT"):
            # Handle Spark's ISO date format that ends with GMT
            try:
                # Remove GMT and parse
                dt_str = value.replace("GMT", "+0000")
                return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError:
                pass
        return value

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ExecutionData":
        """Create an ExecutionData instance from a dictionary."""
        return cls.model_validate(data)


class SparkPlanGraph(BaseModel):
    """Represents a Spark plan graph."""

    nodes: Sequence[Node]
    edges: Sequence[SparkPlanGraphEdge]
    all_nodes: Sequence[Node] = Field([], alias="allNodes")


class SparkPlanGraphNode(BaseModel):
    """Base class for nodes in a Spark plan graph."""

    id: int
    name: str
    metrics: Sequence[Any] = []


class SparkPlanGraphCluster(SparkPlanGraphNode):
    """Represents a cluster of nodes in a Spark plan graph."""

    nodes: Sequence[SparkPlanGraphNode]


# Forward references for type hints
class ExecutorResourceRequest(BaseModel):
    pass


class TaskResourceRequest(BaseModel):
    pass


class ResourceInformation(BaseModel):
    pass


class SparkUI(BaseModel):
    pass


# Update forward references
ShufflePushReadMetrics.model_rebuild()
ShufflePushReadMetricDistributions.model_rebuild()
ExecutorMetricsDistributions.model_rebuild()
ApplicationInfo.model_rebuild()
ApplicationAttemptInfo.model_rebuild()
ResourceProfileInfo.model_rebuild()
ExecutorStageSummary.model_rebuild()
ExecutorSummary.model_rebuild()
JobData.model_rebuild()
RDDStorageInfo.model_rebuild()
RDDDataDistribution.model_rebuild()
StageData.model_rebuild()
TaskData.model_rebuild()
TaskMetrics.model_rebuild()
TaskMetricDistributions.model_rebuild()
ShuffleReadMetricDistributions.model_rebuild()
ApplicationEnvironmentInfo.model_rebuild()



import json
import os
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from mcp.server.fastmcp import FastMCP

from spark_history_mcp.api.emr_persistent_ui_client import EMRPersistentUIClient
from spark_history_mcp.api.spark_client import SparkRestClient
from spark_history_mcp.config.config import Config

from ..utils.utils import ApplicationDiscovery


@dataclass
class AppContext:
    clients: dict[str, SparkRestClient]
    default_client: Optional[SparkRestClient] = None
    app_discovery: Optional[ApplicationDiscovery] = None


class DateTimeEncoder(json.JSONEncoder):
    """Custom JSON encoder that handles datetime objects."""

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:
    # Config() automatically loads from SHS_MCP_CONFIG env var (set in main.py)
    config = Config()

    clients: dict[str, SparkRestClient] = {}
    default_client = None

    for name, server_config in config.servers.items():
        # Check if this is an EMR server configuration
        if server_config.emr_cluster_arn:
            # Create EMR client
            emr_client = EMRPersistentUIClient(server_config)

            # Initialize EMR client (create persistent UI, get presigned URL, setup session)
            base_url, session = emr_client.initialize()

            # Create a modified server config with the base URL
            emr_server_config = server_config.model_copy()
            emr_server_config.url = base_url

            # Create SparkRestClient with the session
            spark_client = SparkRestClient(emr_server_config)
            spark_client.session = session  # Use the authenticated session

            clients[name] = spark_client
        else:
            # Regular Spark REST client
            clients[name] = SparkRestClient(server_config)

        if server_config.default:
            default_client = clients[name]

    app_discovery = ApplicationDiscovery(clients)
    yield AppContext(
        clients=clients, default_client=default_client, app_discovery=app_discovery
    )


def run(config: Config):
    mcp.settings.host = config.mcp.address
    mcp.settings.port = int(config.mcp.port)
    mcp.settings.debug = bool(config.mcp.debug)
    mcp.run(transport=os.getenv("SHS_MCP_TRANSPORT", config.mcp.transports[0]))


mcp = FastMCP("Spark Events", lifespan=app_lifespan)

# Import tools to register them with MCP
from spark_history_mcp.tools import tools  # noqa: E402,F401



"""
Simple timeout handling and parallel execution for MCP tools.
"""

import logging
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Callable, Dict, List, Optional, Tuple

import requests

from spark_history_mcp.api.spark_client import SparkRestClient

logger = logging.getLogger(__name__)


def parallel_execute(
    api_calls: List[Tuple[str, Callable]], max_workers: int = 6, timeout: int = 180
) -> Dict[str, Any]:
    """
    Execute multiple API calls in parallel with error handling.

    Args:
        api_calls: List of (name, function) tuples
        max_workers: Maximum number of parallel workers
        timeout: Total timeout for all operations

    Returns:
        Dictionary with results and errors
    """
    results = {}
    errors = []

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Submit all tasks
        future_to_name = {executor.submit(func): name for name, func in api_calls}

        # Collect results as they complete
        for future in as_completed(future_to_name, timeout=timeout):
            name = future_to_name[future]
            try:
                result = future.result()
                results[name] = result
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 500:
                    # Try to extract the actual error from response text
                    error_text = (
                        e.response.text if hasattr(e.response, "text") else str(e)
                    )
                    if (
                        "OutOfMemoryError" in error_text
                        or "Java heap space" in error_text
                    ):
                        error_msg = f"{name} failed: Spark History Server out of memory (increase SPARK_DAEMON_MEMORY)"
                else:
                    error_msg = (
                        f"{name} failed: HTTP {e.response.status_code} - {str(e)}"
                    )
                errors.append(error_msg)
            except Exception as e:
                error_msg = f"{name} failed: {str(e)}"
                errors.append(error_msg)

    return {"results": results, "errors": errors}


"""
Application discovery with TTL cache and collision handling.
"""


class ApplicationDiscovery:
    def __init__(self, clients: Dict[str, SparkRestClient], ttl: int = 300):
        self.clients = clients
        self.ttl = ttl
        self._cache: Dict[str, Dict] = {}

    def _is_expired(self, entry: Dict) -> bool:
        return time.time() - entry["last_updated"] > self.ttl

    def find_application_servers(self, app_id: str) -> List[str]:
        if app_id in self._cache and not self._is_expired(self._cache[app_id]):
            return self._cache[app_id]["servers"]

        # Search all servers
        servers = []
        for server_name, client in self.clients.items():
            logger.debug(f"Checking for application '{app_id}' in '{server_name}'")
            try:
                client.get_application(app_id)
                servers.append(server_name)
            except Exception as e:
                logger.debug(
                    f"Application '{app_id}' not found on server '{server_name}': {e}"
                )
                continue

        self._cache[app_id] = {"servers": servers, "last_updated": time.time()}

        return servers

    def get_client_for_app(
        self, app_id: str, server_name: Optional[str] = None
    ) -> Tuple[SparkRestClient, str]:
        if server_name:
            if server_name not in self.clients:
                raise ValueError(f"Server '{server_name}' not found")
            return self.clients[server_name], server_name

        servers = self.find_application_servers(app_id)
        if not servers:
            raise ValueError(f"Application '{app_id}' not found on any server")

        # Use first server found
        chosen_server = servers[0]
        return self.clients[chosen_server], chosen_server



import heapq
import logging
from typing import Any, Dict, List, Optional

from spark_history_mcp.core.app import mcp
from spark_history_mcp.models.mcp_types import (
    JobSummary,
    SqlQuerySummary,
)
from spark_history_mcp.models.spark_types import (
    ApplicationInfo,
    ExecutionData,
    JobData,
    JobExecutionStatus,
    SQLExecutionStatus,
    StageData,
    StageStatus,
    TaskMetricDistributions,
)

from ..utils.utils import parallel_execute

logger = logging.getLogger(__name__)


def get_client_or_default(
    ctx, server_name: Optional[str] = None, app_id: Optional[str] = None
):
    """
    Get a client by server name, app discovery, or default client.

    Args:
        ctx: The MCP context
        server_name: Optional server name
        app_id: Optional app ID for discovery

    Returns:
        SparkRestClient: The requested client

    Raises:
        ValueError: If no client is found
    """
    app_discovery = ctx.request_context.lifespan_context.app_discovery
    default_client = ctx.request_context.lifespan_context.default_client

    # If app_id provided, use discovery
    if app_id and not server_name:
        client, _ = app_discovery.get_client_for_app(app_id, server_name)
        return client

    clients = ctx.request_context.lifespan_context.clients

    if server_name:
        client = clients.get(server_name)
        if client:
            return client

    if default_client:
        return default_client

    raise ValueError(
        "No Spark client found. Please specify a valid server name or set a default server."
    )


@mcp.tool()
def list_applications(
    server: Optional[str] = None,
    status: Optional[list[str]] = None,
    min_date: Optional[str] = None,
    max_date: Optional[str] = None,
    min_end_date: Optional[str] = None,
    max_end_date: Optional[str] = None,
    limit: Optional[int] = None,
) -> list[ApplicationInfo]:
    """
    Get a list of applications from the Spark History Server.

    Args:
        server: Optional server name to use (uses default if not specified)
        status: Optional list only applications in the chosen state: [completed|running]
        min_date: Optional earliest start date/time to list
        max_date: Optional latest start date/time to list
        min_end_date: Optional earliest end date/time to list
        max_end_date: Optional latest end date/time to list
        limit: Optional maximum number of applications to return (limits the number of applications listed)
    Date format:
        - Accepted: YYYY-MM-DD["T"HH:mm:ss.SSS"GMT"]
        - Time is optional. If present, it must include milliseconds and the literal GMT.
        - Examples: 2015-02-10, 2015-02-03T16:42:40.000GMT
        - Timezone: All values are interpreted as GMT.
    Returns:
        List of ApplicationInfo objects for all applications
    """
    ctx = mcp.get_context()

    if server:
        # Return from specific server
        client = get_client_or_default(ctx, server)
        return client.list_applications(
            status=status,
            min_date=min_date,
            max_date=max_date,
            min_end_date=min_end_date,
            max_end_date=max_end_date,
            limit=limit,
        )
    else:
        # Return from all servers
        all_apps = []
        clients = ctx.request_context.lifespan_context.clients

        for server_name, client in clients.items():
            try:
                apps = client.list_applications(
                    status=status,
                    min_date=min_date,
                    max_date=max_date,
                    min_end_date=min_end_date,
                    max_end_date=max_end_date,
                    limit=limit,
                )
                all_apps.extend(apps)
            except Exception as e:
                logger.warning(
                    f"Failed to get applications from server '{server_name}': {e}"
                )
                continue  # Skip unreachable servers

        return all_apps


@mcp.tool()
def get_application(app_id: str, server: Optional[str] = None) -> ApplicationInfo:
    """
    Get detailed information about a specific Spark application.

    Retrieves comprehensive information about a Spark application including its
    status, resource usage, duration, and attempt details.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        ApplicationInfo object containing application details
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    return client.get_application(app_id)


@mcp.tool()
def list_jobs(
    app_id: str, server: Optional[str] = None, status: Optional[list[str]] = None
) -> list:
    """
    Get a list of all jobs for a Spark application.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        status: Optional list of job status values to filter by

    Returns:
        List of JobData objects for the application
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Convert string status values to JobExecutionStatus enum if provided
    job_statuses = None
    if status:
        job_statuses = [JobExecutionStatus.from_string(s) for s in status]

    return client.list_jobs(app_id=app_id, status=job_statuses)


@mcp.tool()
def list_slowest_jobs(
    app_id: str,
    server: Optional[str] = None,
    include_running: bool = False,
    n: int = 5,
) -> List[JobData]:
    """
    Get the N slowest jobs for a Spark application.

    Retrieves all jobs for the application and returns the ones with the longest duration.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        include_running: Whether to include running jobs in the search
        n: Number of slowest jobs to return (default: 5)

    Returns:
        List of JobData objects for the slowest jobs, or empty list if no jobs found
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Get all jobs
    jobs = client.list_jobs(app_id=app_id)

    if not jobs:
        return []

    # Filter out running jobs if not included
    if not include_running:
        jobs = [job for job in jobs if job.status != JobExecutionStatus.RUNNING.value]

    if not jobs:
        return []

    def get_job_duration(job):
        if job.completion_time and job.submission_time:
            return (job.completion_time - job.submission_time).total_seconds()
        return 0

    return heapq.nlargest(n, jobs, key=get_job_duration)


@mcp.tool()
def list_stages(
    app_id: str,
    server: Optional[str] = None,
    status: Optional[list[str]] = None,
    with_summaries: bool = False,
) -> list:
    """
    Get a list of all stages for a Spark application.

    Retrieves information about stages in a Spark application with options to filter
    by status and include additional details and summary metrics.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        status: Optional list of stage status values to filter by
        with_summaries: Whether to include summary metrics in the response

    Returns:
        List of StageData objects for the application
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Convert string status values to StageStatus enum if provided
    stage_statuses = None
    if status:
        stage_statuses = [StageStatus.from_string(s) for s in status]

    return client.list_stages(
        app_id=app_id,
        status=stage_statuses,
        with_summaries=with_summaries,
    )


@mcp.tool()
def list_slowest_stages(
    app_id: str,
    server: Optional[str] = None,
    include_running: bool = False,
    n: int = 5,
) -> List[StageData]:
    """
    Get the N slowest stages for a Spark application.

    Retrieves all stages for the application and returns the ones with the longest duration.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        include_running: Whether to include running stages in the search
        n: Number of slowest stages to return (default: 5)

    Returns:
        List of StageData objects for the slowest stages, or empty list if no stages found
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    stages = client.list_stages(app_id=app_id)

    # Filter out running stages if not included. This avoids using the `details` param which can significantly slow down the execution time
    if not include_running:
        stages = [stage for stage in stages if stage.status != "RUNNING"]

    if not stages:
        return []

    def get_stage_duration(stage: StageData):
        if stage.completion_time and stage.first_task_launched_time:
            return (
                stage.completion_time - stage.first_task_launched_time
            ).total_seconds()
        return 0

    return heapq.nlargest(n, stages, key=get_stage_duration)


@mcp.tool()
def get_stage(
    app_id: str,
    stage_id: int,
    attempt_id: Optional[int] = None,
    server: Optional[str] = None,
    with_summaries: bool = False,
) -> StageData:
    """
    Get information about a specific stage.

    Args:
        app_id: The Spark application ID
        stage_id: The stage ID
        attempt_id: Optional stage attempt ID (if not provided, returns the latest attempt)
        server: Optional server name to use (uses default if not specified)
        with_summaries: Whether to include summary metrics

    Returns:
        StageData object containing stage information
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    if attempt_id is not None:
        # Get specific attempt
        stage_data = client.get_stage_attempt(
            app_id=app_id,
            stage_id=stage_id,
            attempt_id=attempt_id,
            details=False,
            with_summaries=with_summaries,
        )
    else:
        # Get all attempts and use the latest one
        stages = client.list_stage_attempts(
            app_id=app_id,
            stage_id=stage_id,
            details=False,
            with_summaries=with_summaries,
        )

        if not stages:
            raise ValueError(f"No stage found with ID {stage_id}")

        # If multiple attempts exist, get the one with the highest attempt_id
        if isinstance(stages, list):
            stage_data = max(stages, key=lambda s: s.attempt_id)
        else:
            stage_data = stages

    # If summaries were requested but metrics distributions are missing, fetch them separately
    if with_summaries and (
        not hasattr(stage_data, "task_metrics_distributions")
        or stage_data.task_metrics_distributions is None
    ):
        task_summary = client.get_stage_task_summary(
            app_id=app_id,
            stage_id=stage_id,
            attempt_id=stage_data.attempt_id,
        )
        stage_data.task_metrics_distributions = task_summary

    return stage_data


@mcp.tool()
def get_environment(app_id: str, server: Optional[str] = None):
    """
    Get the comprehensive Spark runtime configuration for a Spark application.

    Details including JVM information, Spark properties, system properties,
    classpath entries, and environment variables.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        ApplicationEnvironmentInfo object containing environment details
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    return client.get_environment(app_id=app_id)


@mcp.tool()
def list_executors(
    app_id: str, server: Optional[str] = None, include_inactive: bool = False
):
    """
    Get executor information for a Spark application.

    Retrieves a list of executors (active by default) for the specified Spark application
    with their resource allocation, task statistics, and performance metrics.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        include_inactive: Whether to include inactive executors (default: False)

    Returns:
        List of ExecutorSummary objects containing executor information
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    if include_inactive:
        return client.list_all_executors(app_id=app_id)
    else:
        return client.list_executors(app_id=app_id)


@mcp.tool()
def get_executor(app_id: str, executor_id: str, server: Optional[str] = None):
    """
    Get information about a specific executor.

    Retrieves detailed information about a single executor including resource allocation,
    task statistics, memory usage, and performance metrics.

    Args:
        app_id: The Spark application ID
        executor_id: The executor ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        ExecutorSummary object containing executor details or None if not found
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Get all executors and find the one with matching ID
    executors = client.list_all_executors(app_id=app_id)

    for executor in executors:
        if executor.id == executor_id:
            return executor

    return None


@mcp.tool()
def get_executor_summary(app_id: str, server: Optional[str] = None):
    """
    Aggregates metrics across all executors for a Spark application.

    Retrieves all executors (active and inactive) and calculates summary statistics
    including memory usage, disk usage, task counts, and performance metrics.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        Dictionary containing aggregated executor metrics
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    executors = client.list_all_executors(app_id=app_id)
    return _calculate_executor_metrics(executors)


@mcp.tool()
def compare_job_environments(
    app_id1: str, app_id2: str, server: Optional[str] = None
) -> Dict[str, Any]:
    """
    Compare Spark environment configurations between two jobs.

    Identifies differences in Spark properties, JVM settings, system properties,
    and other configuration parameters between two Spark applications.

    Args:
        app_id1: First Spark application ID
        app_id2: Second Spark application ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        Dictionary containing configuration differences and similarities
    """
    ctx = mcp.get_context()
    client1 = get_client_or_default(ctx, server, app_id1)
    client2 = get_client_or_default(ctx, server, app_id2)

    env1 = client1.get_environment(app_id=app_id1)
    env2 = client2.get_environment(app_id=app_id2)

    def props_to_dict(props):
        return {k: v for k, v in props} if props else {}

    spark_props1 = props_to_dict(env1.spark_properties)
    spark_props2 = props_to_dict(env2.spark_properties)

    system_props1 = props_to_dict(env1.system_properties)
    system_props2 = props_to_dict(env2.system_properties)

    comparison = {
        "applications": {"app1": app_id1, "app2": app_id2},
        "runtime_comparison": {
            "app1": {
                "java_version": env1.runtime.java_version,
                "java_home": env1.runtime.java_home,
                "scala_version": env1.runtime.scala_version,
            },
            "app2": {
                "java_version": env2.runtime.java_version,
                "java_home": env2.runtime.java_home,
                "scala_version": env2.runtime.scala_version,
            },
        },
        "spark_properties": {
            "common": {
                k: {"app1": v, "app2": spark_props2.get(k)}
                for k, v in spark_props1.items()
                if k in spark_props2 and v == spark_props2[k]
            },
            "different": {
                k: {"app1": v, "app2": spark_props2.get(k, "NOT_SET")}
                for k, v in spark_props1.items()
                if k in spark_props2 and v != spark_props2[k]
            },
            "only_in_app1": {
                k: v for k, v in spark_props1.items() if k not in spark_props2
            },
            "only_in_app2": {
                k: v for k, v in spark_props2.items() if k not in spark_props1
            },
        },
        "system_properties": {
            "key_differences": {
                k: {
                    "app1": system_props1.get(k, "NOT_SET"),
                    "app2": system_props2.get(k, "NOT_SET"),
                }
                for k in [
                    "java.version",
                    "java.runtime.version",
                    "os.name",
                    "os.version",
                    "user.timezone",
                    "file.encoding",
                ]
                if system_props1.get(k) != system_props2.get(k)
            }
        },
    }

    return comparison


def _calculate_executor_metrics(executors):
    """Calculate executor summary metrics from executor list."""
    return {
        "total_executors": len(executors),
        "active_executors": sum(1 for e in executors if e.is_active),
        "memory_used": sum(
            e.memory_metrics.used_on_heap_storage_memory
            + e.memory_metrics.used_off_heap_storage_memory
            for e in executors
        ),
        "disk_used": sum(e.disk_used for e in executors),
        "completed_tasks": sum(e.completed_tasks for e in executors),
        "failed_tasks": sum(e.failed_tasks for e in executors),
        "total_duration": sum(e.total_duration for e in executors),
        "total_gc_time": sum(e.total_gc_time for e in executors),
        "total_input_bytes": sum(e.total_input_bytes for e in executors),
        "total_shuffle_read": sum(e.total_shuffle_read for e in executors),
        "total_shuffle_write": sum(e.total_shuffle_write for e in executors),
    }


def _calc_executor_summary_from_client(client, app_id: str):
    """Helper function to calculate executor summary without MCP context."""
    executors = client.list_all_executors(app_id=app_id)
    return _calculate_executor_metrics(executors)


@mcp.tool()
def compare_job_performance(
    app_id1: str, app_id2: str, server: Optional[str] = None
) -> Dict[str, Any]:
    """
    Compare performance metrics between two Spark jobs.

    Analyzes execution times, resource usage, task distribution, and other
    performance indicators to identify differences between jobs.

    Args:
        app_id1: First Spark application ID
        app_id2: Second Spark application ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        Dictionary containing detailed performance comparison
    """
    ctx = mcp.get_context()
    client1 = get_client_or_default(ctx, server, app_id1)
    client2 = get_client_or_default(ctx, server, app_id2)

    # Define API calls for parallel execution
    api_calls = [
        ("app1", lambda: client1.get_application(app_id1)),
        ("app2", lambda: client2.get_application(app_id2)),
        ("exec_summary1", lambda: _calc_executor_summary_from_client(client1, app_id1)),
        ("exec_summary2", lambda: _calc_executor_summary_from_client(client2, app_id2)),
        ("jobs1", lambda: client1.list_jobs(app_id=app_id1)),
        ("jobs2", lambda: client2.list_jobs(app_id=app_id2)),
    ]

    # Execute all API calls in parallel
    execution_result = parallel_execute(
        api_calls,
        max_workers=6,
        timeout=300,  # Apply generous timeout for large scale Spark applications
    )

    # If parallel execution fails, try sequential as fallback
    if execution_result["errors"] and len(execution_result["results"]) == 0:
        try:
            # Sequential fallback - get basic info first
            app1 = client1.get_application(app_id1)
            app2 = client2.get_application(app_id2)

            # Use the actual errors from parallel execution
            error_summary = "; ".join(execution_result["errors"])
            return {
                "error": f"Parallel execution failed: {error_summary}. Falling back to basic app info only.",
                "partial_data": {
                    "app1": {"id": app_id1, "name": app1.name},
                    "app2": {"id": app_id2, "name": app2.name},
                },
            }
        except Exception as e:
            # If even basic app info fails, provide the original errors plus this failure
            all_errors = execution_result["errors"] + [
                f"Sequential fallback failed: {str(e)}"
            ]
            return {"error": f"Complete failure: {'; '.join(all_errors)}"}

    if execution_result["errors"]:
        return {"error": f"API failures: {'; '.join(execution_result['errors'])}"}

    results = execution_result["results"]

    # Extract results
    app1 = results["app1"]
    app2 = results["app2"]
    exec_summary1 = results["exec_summary1"]
    exec_summary2 = results["exec_summary2"]
    jobs1 = results["jobs1"]
    jobs2 = results["jobs2"]

    # Calculate job duration statistics
    def calc_job_stats(jobs):
        if not jobs:
            return {"count": 0, "total_duration": 0, "avg_duration": 0}

        completed_jobs = [j for j in jobs if j.completion_time and j.submission_time]
        if not completed_jobs:
            return {"count": len(jobs), "total_duration": 0, "avg_duration": 0}

        durations = [
            (j.completion_time - j.submission_time).total_seconds()
            for j in completed_jobs
        ]

        return {
            "count": len(jobs),
            "completed_count": len(completed_jobs),
            "total_duration": sum(durations),
            "avg_duration": sum(durations) / len(durations),
            "min_duration": min(durations),
            "max_duration": max(durations),
        }

    job_stats1 = calc_job_stats(jobs1)
    job_stats2 = calc_job_stats(jobs2)

    comparison = {
        "applications": {
            "app1": {"id": app_id1, "name": app1.name},
            "app2": {"id": app_id2, "name": app2.name},
        },
        "resource_allocation": {
            "app1": {
                "cores_granted": app1.cores_granted,
                "max_cores": app1.max_cores,
                "cores_per_executor": app1.cores_per_executor,
                "memory_per_executor_mb": app1.memory_per_executor_mb,
            },
            "app2": {
                "cores_granted": app2.cores_granted,
                "max_cores": app2.max_cores,
                "cores_per_executor": app2.cores_per_executor,
                "memory_per_executor_mb": app2.memory_per_executor_mb,
            },
        },
        "executor_metrics": {
            "app1": exec_summary1,
            "app2": exec_summary2,
            "comparison": {
                "executor_count_ratio": exec_summary2["total_executors"]
                / max(exec_summary1["total_executors"], 1),
                "memory_usage_ratio": exec_summary2["memory_used"]
                / max(exec_summary1["memory_used"], 1),
                "task_completion_ratio": exec_summary2["completed_tasks"]
                / max(exec_summary1["completed_tasks"], 1),
                "gc_time_ratio": exec_summary2["total_gc_time"]
                / max(exec_summary1["total_gc_time"], 1),
            },
        },
        "job_performance": {
            "app1": job_stats1,
            "app2": job_stats2,
            "comparison": {
                "job_count_ratio": job_stats2["count"] / max(job_stats1["count"], 1),
                "avg_duration_ratio": job_stats2["avg_duration"]
                / max(job_stats1["avg_duration"], 1)
                if job_stats1["avg_duration"] > 0
                else 0,
                "total_duration_ratio": job_stats2["total_duration"]
                / max(job_stats1["total_duration"], 1)
                if job_stats1["total_duration"] > 0
                else 0,
            },
        },
    }

    return comparison


@mcp.tool()
def compare_sql_execution_plans(
    app_id1: str,
    app_id2: str,
    execution_id1: Optional[int] = None,
    execution_id2: Optional[int] = None,
    server: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Compare SQL execution plans between two Spark jobs.

    Analyzes the logical and physical plans, identifies differences in operations,
    and compares execution metrics between SQL queries.

    Args:
        app_id1: First Spark application ID
        app_id2: Second Spark application ID
        execution_id1: Optional specific execution ID for first app (uses longest if not specified)
        execution_id2: Optional specific execution ID for second app (uses longest if not specified)
        server: Optional server name to use (uses default if not specified)

    Returns:
        Dictionary containing SQL execution plan comparison
    """
    ctx = mcp.get_context()
    client1 = get_client_or_default(ctx, server, app_id1)
    client2 = get_client_or_default(ctx, server, app_id2)

    # Get SQL executions for both applications
    sql_execs1 = client1.get_sql_list(
        app_id=app_id1, details=True, plan_description=True
    )
    sql_execs2 = client2.get_sql_list(
        app_id=app_id2, details=True, plan_description=True
    )

    # If specific execution IDs not provided, use the longest running ones
    if execution_id1 is None and sql_execs1:
        execution_id1 = max(sql_execs1, key=lambda x: x.duration or 0).id
    if execution_id2 is None and sql_execs2:
        execution_id2 = max(sql_execs2, key=lambda x: x.duration or 0).id

    if execution_id1 is None or execution_id2 is None:
        return {
            "error": "No SQL executions found in one or both applications",
            "app1_sql_count": len(sql_execs1),
            "app2_sql_count": len(sql_execs2),
        }

    # Get specific execution details
    exec1 = client1.get_sql_execution(
        app_id1, execution_id1, details=True, plan_description=True
    )
    exec2 = client2.get_sql_execution(
        app_id2, execution_id2, details=True, plan_description=True
    )

    # Analyze nodes and operations
    def analyze_nodes(execution):
        node_types = {}
        for node in execution.nodes:
            node_type = node.node_name
            if node_type not in node_types:
                node_types[node_type] = 0
            node_types[node_type] += 1
        return node_types

    nodes1 = analyze_nodes(exec1)
    nodes2 = analyze_nodes(exec2)

    all_node_types = set(nodes1.keys()) | set(nodes2.keys())

    comparison = {
        "applications": {"app1": app_id1, "app2": app_id2},
        "executions": {
            "app1": {
                "execution_id": execution_id1,
                "duration": exec1.duration,
                "status": exec1.status,
                "node_count": len(exec1.nodes),
                "edge_count": len(exec1.edges),
            },
            "app2": {
                "execution_id": execution_id2,
                "duration": exec2.duration,
                "status": exec2.status,
                "node_count": len(exec2.nodes),
                "edge_count": len(exec2.edges),
            },
        },
        "plan_structure": {
            "node_type_comparison": {
                node_type: {
                    "app1_count": nodes1.get(node_type, 0),
                    "app2_count": nodes2.get(node_type, 0),
                }
                for node_type in sorted(all_node_types)
            },
            "complexity_metrics": {
                "node_count_ratio": len(exec2.nodes) / max(len(exec1.nodes), 1),
                "edge_count_ratio": len(exec2.edges) / max(len(exec1.edges), 1),
                "duration_ratio": (exec2.duration or 0) / max(exec1.duration or 1, 1),
            },
        },
        "job_associations": {
            "app1": {
                "running_jobs": exec1.running_job_ids,
                "success_jobs": exec1.success_job_ids,
                "failed_jobs": exec1.failed_job_ids,
            },
            "app2": {
                "running_jobs": exec2.running_job_ids,
                "success_jobs": exec2.success_job_ids,
                "failed_jobs": exec2.failed_job_ids,
            },
        },
    }

    return comparison


@mcp.tool()
def get_stage_task_summary(
    app_id: str,
    stage_id: int,
    attempt_id: int = 0,
    server: Optional[str] = None,
    quantiles: str = "0.05,0.25,0.5,0.75,0.95",
) -> TaskMetricDistributions:
    """
    Get a summary of task metrics for a specific stage.

    Retrieves statistical distributions of task metrics for a stage, including
    execution times, memory usage, I/O metrics, and shuffle metrics.

    Args:
        app_id: The Spark application ID
        stage_id: The stage ID
        attempt_id: The stage attempt ID (default: 0)
        server: Optional server name to use (uses default if not specified)
        quantiles: Comma-separated list of quantiles to use for summary metrics

    Returns:
        TaskMetricDistributions object containing metric distributions
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    return client.get_stage_task_summary(
        app_id=app_id, stage_id=stage_id, attempt_id=attempt_id, quantiles=quantiles
    )


def truncate_plan_description(plan_desc: str, max_length: int) -> str:
    """
    Truncate plan description while preserving structure.

    Args:
        plan_desc: The plan description to truncate
        max_length: Maximum length in characters

    Returns:
        Truncated plan description with indicator if truncated
    """
    if not plan_desc or len(plan_desc) <= max_length:
        return plan_desc

    # Try to truncate at a logical boundary (end of a line)
    truncated = plan_desc[:max_length]
    last_newline = truncated.rfind("\n")

    # If we can preserve most content by truncating at newline, do so
    if last_newline > max_length * 0.8:
        truncated = truncated[:last_newline]

    return truncated + "\n... [truncated]"


@mcp.tool()
def list_slowest_sql_queries(
    app_id: str,
    server: Optional[str] = None,
    attempt_id: Optional[str] = None,
    top_n: int = 1,
    page_size: int = 100,
    include_running: bool = False,
    include_plan_description: Optional[bool] = None,
    plan_description_max_length: int = 2000,
) -> List[SqlQuerySummary]:
    """
    Get the N slowest SQL queries for a Spark application.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        attempt_id: Optional attempt ID
        top_n: Number of slowest queries to return (default: 1)
        page_size: Number of executions to fetch per page (default: 100)
        include_running: Whether to include running queries (default: False)
        include_plan_description: Whether to include execution plans (uses server config if not specified)
        plan_description_max_length: Max characters for plan description (default: 1500)

    Returns:
        List of SqlQuerySummary objects for the slowest queries
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Config takes priority: if config is set (True/False), use it; otherwise default to True
    if client.config.include_plan_description is not None:
        include_plan_description = client.config.include_plan_description
    else:
        include_plan_description = True

    all_executions: List[ExecutionData] = []
    offset = 0

    # Fetch all pages of SQL executions
    while True:
        executions: List[ExecutionData] = client.get_sql_list(
            app_id=app_id,
            attempt_id=attempt_id,
            details=True,
            plan_description=True,
            offset=offset,
            length=page_size,
        )

        if not executions:
            break

        all_executions.extend(executions)
        offset += page_size

        # If we got fewer executions than the page size, we've reached the end
        if len(executions) < page_size:
            break

    # Filter out running queries if not included
    if not include_running:
        all_executions = [
            e for e in all_executions if e.status != SQLExecutionStatus.RUNNING.value
        ]

    # Get the top N slowest executions
    slowest_executions = heapq.nlargest(top_n, all_executions, key=lambda e: e.duration)

    # Create simplified results without additional API calls. Raw object is too verbose.
    simplified_results = []
    for execution in slowest_executions:
        job_summary = JobSummary(
            success_job_ids=execution.success_job_ids,
            failed_job_ids=execution.failed_job_ids,
            running_job_ids=execution.running_job_ids,
        )

        # Handle plan description based on include_plan_description flag
        plan_description = ""
        if include_plan_description and execution.plan_description:
            plan_description = truncate_plan_description(
                execution.plan_description, plan_description_max_length
            )

        query_summary = SqlQuerySummary(
            id=execution.id,
            duration=execution.duration,
            description=execution.description,
            status=execution.status,
            submission_time=execution.submission_time.isoformat()
            if execution.submission_time
            else None,
            plan_description=plan_description,
            job_summary=job_summary,
        )

        simplified_results.append(query_summary)

    return simplified_results


@mcp.tool()
def get_job_bottlenecks(
    app_id: str, server: Optional[str] = None, top_n: int = 5
) -> Dict[str, Any]:
    """
    Identify performance bottlenecks in a Spark job.

    Analyzes stages, tasks, and executors to find the most time-consuming
    operations and resource-intensive components.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)
        top_n: Number of top bottlenecks to return

    Returns:
        Dictionary containing identified bottlenecks and recommendations
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Get slowest stages
    slowest_stages = list_slowest_stages(app_id, server, False, top_n)

    # Get slowest jobs
    slowest_jobs = list_slowest_jobs(app_id, server, False, top_n)

    # Get executor summary
    exec_summary = get_executor_summary(app_id, server)

    all_stages = client.list_stages(app_id=app_id)

    # Identify stages with high spill
    high_spill_stages = []
    for stage in all_stages:
        if (
            stage.memory_bytes_spilled
            and stage.memory_bytes_spilled > 100 * 1024 * 1024
        ):  # > 100MB
            high_spill_stages.append(
                {
                    "stage_id": stage.stage_id,
                    "attempt_id": stage.attempt_id,
                    "name": stage.name,
                    "memory_spilled_mb": stage.memory_bytes_spilled / (1024 * 1024),
                    "disk_spilled_mb": stage.disk_bytes_spilled / (1024 * 1024)
                    if stage.disk_bytes_spilled
                    else 0,
                }
            )

    high_spill_stages = heapq.nlargest(
        len(high_spill_stages), high_spill_stages, key=lambda x: x["memory_spilled_mb"]
    )

    # Identify GC pressure
    gc_pressure = (
        exec_summary["total_gc_time"] / max(exec_summary["total_duration"], 1)
        if exec_summary["total_duration"] > 0
        else 0
    )

    bottlenecks = {
        "application_id": app_id,
        "performance_bottlenecks": {
            "slowest_stages": [
                {
                    "stage_id": stage.stage_id,
                    "attempt_id": stage.attempt_id,
                    "name": stage.name,
                    "duration_seconds": (
                        stage.completion_time - stage.submission_time
                    ).total_seconds()
                    if stage.completion_time and stage.submission_time
                    else 0,
                    "task_count": stage.num_tasks,
                    "failed_tasks": stage.num_failed_tasks,
                }
                for stage in slowest_stages[:top_n]
            ],
            "slowest_jobs": [
                {
                    "job_id": job.job_id,
                    "name": job.name,
                    "duration_seconds": (
                        job.completion_time - job.submission_time
                    ).total_seconds()
                    if job.completion_time and job.submission_time
                    else 0,
                    "failed_tasks": job.num_failed_tasks,
                    "status": job.status,
                }
                for job in slowest_jobs[:top_n]
            ],
        },
        "resource_bottlenecks": {
            "memory_spill_stages": high_spill_stages[:top_n],
            "gc_pressure_ratio": gc_pressure,
            "executor_utilization": {
                "total_executors": exec_summary["total_executors"],
                "active_executors": exec_summary["active_executors"],
                "utilization_ratio": exec_summary["active_executors"]
                / max(exec_summary["total_executors"], 1),
            },
        },
        "recommendations": [],
    }

    # Generate recommendations
    if gc_pressure > 0.1:  # More than 10% time in GC
        bottlenecks["recommendations"].append(
            {
                "type": "memory",
                "priority": "high",
                "issue": f"High GC pressure ({gc_pressure:.1%})",
                "suggestion": "Consider increasing executor memory or reducing memory usage",
            }
        )

    if high_spill_stages:
        bottlenecks["recommendations"].append(
            {
                "type": "memory",
                "priority": "high",
                "issue": f"Memory spilling detected in {len(high_spill_stages)} stages",
                "suggestion": "Increase executor memory or optimize data partitioning",
            }
        )

    if exec_summary["failed_tasks"] > 0:
        bottlenecks["recommendations"].append(
            {
                "type": "reliability",
                "priority": "medium",
                "issue": f"{exec_summary['failed_tasks']} failed tasks",
                "suggestion": "Investigate task failures and consider increasing task retry settings",
            }
        )

    return bottlenecks


@mcp.tool()
def get_resource_usage_timeline(
    app_id: str, server: Optional[str] = None
) -> Dict[str, Any]:
    """
    Get resource usage timeline for a Spark application.

    Provides a chronological view of resource allocation and usage patterns
    including executor additions/removals and stage execution overlap.

    Args:
        app_id: The Spark application ID
        server: Optional server name to use (uses default if not specified)

    Returns:
        Dictionary containing timeline of resource usage
    """
    ctx = mcp.get_context()
    client = get_client_or_default(ctx, server, app_id)

    # Get application info
    app = client.get_application(app_id)

    # Get all executors
    executors = client.list_all_executors(app_id=app_id)

    # Get stages
    stages = client.list_stages(app_id=app_id)

    # Create timeline events
    timeline_events = []

    # Add executor events
    for executor in executors:
        if executor.add_time:
            timeline_events.append(
                {
                    "timestamp": executor.add_time,
                    "type": "executor_add",
                    "executor_id": executor.id,
                    "cores": executor.total_cores,
                    "memory_mb": executor.max_memory / (1024 * 1024)
                    if executor.max_memory
                    else 0,
                }
            )

        if executor.remove_time:
            timeline_events.append(
                {
                    "timestamp": executor.remove_time,
                    "type": "executor_remove",
                    "executor_id": executor.id,
                    "reason": executor.remove_reason,
                }
            )

    # Add stage events
    for stage in stages:
        if stage.submission_time:
            timeline_events.append(
                {
                    "timestamp": stage.submission_time,
                    "type": "stage_start",
                    "stage_id": stage.stage_id,
                    "attempt_id": stage.attempt_id,
                    "name": stage.name,
                    "task_count": stage.num_tasks,
                }
            )

        if stage.completion_time:
            timeline_events.append(
                {
                    "timestamp": stage.completion_time,
                    "type": "stage_end",
                    "stage_id": stage.stage_id,
                    "attempt_id": stage.attempt_id,
                    "status": stage.status,
                    "duration_seconds": (
                        stage.completion_time - stage.submission_time
                    ).total_seconds()
                    if stage.submission_time
                    else 0,
                }
            )

    # Sort events by timestamp
    timeline_events.sort(key=lambda x: x["timestamp"])

    # Calculate resource utilization over time
    active_executors = 0
    total_cores = 0
    total_memory = 0

    resource_timeline = []

    for event in timeline_events:
        if event["type"] == "executor_add":
            active_executors += 1
            total_cores += event["cores"]
            total_memory += event["memory_mb"]
        elif event["type"] == "executor_remove":
            active_executors -= 1
            # Note: We don't have cores/memory info in remove events

        resource_timeline.append(
            {
                "timestamp": event["timestamp"],
                "active_executors": active_executors,
                "total_cores": total_cores,
                "total_memory_mb": total_memory,
                "event": event,
            }
        )

    return {
        "application_id": app_id,
        "application_name": app.name,
        "summary": {
            "total_events": len(timeline_events),
            "executor_additions": len(
                [e for e in timeline_events if e["type"] == "executor_add"]
            ),
            "executor_removals": len(
                [e for e in timeline_events if e["type"] == "executor_remove"]
            ),
            "stage_executions": len(
                [e for e in timeline_events if e["type"] == "stage_start"]
            ),
            "peak_executors": max(
                [r["active_executors"] for r in resource_timeline] + [0]
            ),
            "peak_cores": max([r["total_cores"] for r in resource_timeline] + [0]),
        },
    }



import re
from typing import Any, Dict, List, Optional, Type, TypeVar
from urllib.parse import urljoin

import requests
from pydantic import BaseModel

from spark_history_mcp.config.config import ServerConfig
from spark_history_mcp.models.spark_types import (
    ApplicationAttemptInfo,
    ApplicationEnvironmentInfo,
    ApplicationInfo,
    ExecutionData,
    ExecutorSummary,
    JobData,
    JobExecutionStatus,
    ProcessSummary,
    RDDStorageInfo,
    StageData,
    StageStatus,
    TaskData,
    TaskMetricDistributions,
    TaskStatus,
    ThreadStackTrace,
    VersionInfo,
)

T = TypeVar("T", bound=BaseModel)


class SparkRestClient:
    """
    Python client for the Spark REST API.
    """

    def __init__(self, server_config: ServerConfig):
        """
        Initialize the Spark REST client.

        Args:
            server_config: Configuration object
        """
        self.config = server_config
        self.base_url = self.config.url.rstrip("/") + "/api/v1"
        self.auth = None
        self.session = None
        self.use_proxy = self.config.use_proxy
        self.proxies = (
            self.use_proxy
            and {
                "http": "socks5h://localhost:8157",
                "https": "socks5h://localhost:8157",
            }
            or None
        )
        self.pattern = re.compile(r"(.*?/applications/[^/]+/)(.+)")

        # Determine whether to verify SSL certificates and timeout
        # Default to True for verify_ssl and 30 seconds for timeout if not specified
        self.verify_ssl = self.config.verify_ssl
        self.timeout = self.config.timeout

        # Set up authentication if provided
        if self.config.auth:
            if self.config.auth.username and self.config.auth.password:
                self.auth = (self.config.auth.username, self.config.auth.password)

    def _make_request(
        self, request_url: str, params: Optional[Dict[str, Any]]
    ) -> requests.Response:
        """
        Make a GET request to the Spark REST API.

        Args:
            request_url: The request URL
            params: Optional query parameters

        Returns:
            The response from the API
        """
        headers = {"Accept": "application/json"}

        # Add token to headers if provided
        if self.config.auth and self.config.auth.token:
            headers["Authorization"] = f"Bearer {self.config.auth.token}"

        # Use the verify_ssl setting for HTTPS requests
        verify = self.verify_ssl

        # Use the session if available, otherwise use requests directly
        if self.session:
            # Add headers to the session
            for key, value in headers.items():
                self.session.headers[key] = value

            response = self.session.get(
                request_url,
                params=params,
                timeout=self.timeout,
                verify=verify,
                proxies=self.proxies,
            )
        else:
            response = requests.get(
                request_url,
                params=params,
                headers=headers,
                auth=self.auth,
                timeout=self.timeout,
                verify=verify,
                proxies=self.proxies,
            )
        return response

    def _modify_url(self, url):
        match = self.pattern.search(url)
        if match:
            prefix = match.group(1)
            suffix = match.group(2)
            # Check if the suffix already starts with a number (attempt ID)
            if not re.match(r"^\d+/", suffix):
                # If no attempt ID present, add the first (and probably only) attempt of the app running on YARN
                app_attempt_id = 1
                return f"{prefix}{app_attempt_id}/{suffix}"
        return url

    def _get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """
        Make a GET request to the Spark REST API.

        Args:
            endpoint: The API endpoint to call
            params: Optional query parameters

        Returns:
            The JSON response from the API
        """
        url = urljoin(self.base_url + "/", endpoint.lstrip("/"))

        try:
            # Try original URL first
            first_response = self._make_request(url, params)
            first_response.raise_for_status()
            return first_response.json()
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404 and "/applications/" in url:
                modified_url = self._modify_url(url)
                try:
                    second_response = self._make_request(modified_url, params)
                    second_response.raise_for_status()
                    return second_response.json()
                except requests.exceptions.HTTPError as e2:
                    raise e2 from e  # Chain the exception with the original error
            # Raise the original error
            raise e from None

    def _parse_model(self, data: Dict[str, Any], model_class: Type[T]) -> T:
        """
        Parse JSON data into a Pydantic model.

        Args:
            data: The JSON data to parse
            model_class: The Pydantic model class to use

        Returns:
            An instance of the model class
        """
        return model_class.model_validate(data)

    def _parse_model_list(
        self, data: List[Dict[str, Any]], model_class: Type[T]
    ) -> List[T]:
        """
        Parse a list of JSON data into a list of Pydantic models.

        Args:
            data: The list of JSON data to parse
            model_class: The Pydantic model class to use

        Returns:
            A list of instances of the model class
        """
        return [self._parse_model(item, model_class) for item in data]

    def get_version(self) -> VersionInfo:
        """Get the Spark version."""
        data = self._get("version")
        return self._parse_model(data, VersionInfo)

    def list_applications(
        self,
        status: Optional[List[str]] = None,
        min_date: Optional[str] = None,
        max_date: Optional[str] = None,
        min_end_date: Optional[str] = None,
        max_end_date: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> List[ApplicationInfo]:
        """
        Get a list of all applications.

        Args:
            status: Filter by application status (COMPLETED, RUNNING)
            min_date: Minimum start date (yyyy-MM-dd'T'HH:mm:ss.SSSz or yyyy-MM-dd)
            max_date: Maximum start date
            min_end_date: Minimum end date
            max_end_date: Maximum end date
            limit: Maximum number of applications to return

        Returns:
            List of ApplicationInfo objects
        """
        params = {}
        if status:
            params["status"] = status
        if min_date:
            params["minDate"] = min_date
        if max_date:
            params["maxDate"] = max_date
        if min_end_date:
            params["minEndDate"] = min_end_date
        if max_end_date:
            params["maxEndDate"] = max_end_date
        if limit:
            params["limit"] = limit

        data = self._get("applications", params)
        return self._parse_model_list(data, ApplicationInfo)

    def get_application(self, app_id: str) -> ApplicationInfo:
        """
        Get information about a specific application.

        Args:
            app_id: The application ID

        Returns:
            ApplicationInfo object
        """
        data = self._get(f"applications/{app_id}")
        return self._parse_model(data, ApplicationInfo)

    def get_application_attempt(
        self, app_id: str, attempt_id: str
    ) -> ApplicationAttemptInfo:
        """
        Get information about a specific application attempt.

        Args:
            app_id: The application ID
            attempt_id: The attempt ID

        Returns:
            ApplicationAttemptInfo object
        """
        data = self._get(f"applications/{app_id}/{attempt_id}")
        return self._parse_model(data, ApplicationAttemptInfo)

    def list_jobs(
        self, app_id: str, status: Optional[List[JobExecutionStatus]] = None
    ) -> List[JobData]:
        """
        Get a list of all jobs for an application.

        Args:
            app_id: The application ID
            status: Filter by job status

        Returns:
            List of JobData objects
        """
        params = {}
        if status:
            params["status"] = [s.value for s in status]

        data = self._get(f"applications/{app_id}/jobs", params)
        return self._parse_model_list(data, JobData)

    def get_job(self, app_id: str, job_id: int) -> JobData:
        """
        Get information about a specific job.

        Args:
            app_id: The application ID
            job_id: The job ID

        Returns:
            JobData object
        """
        data = self._get(f"applications/{app_id}/jobs/{job_id}")
        return self._parse_model(data, JobData)

    def list_stages(
        self,
        app_id: str,
        status: Optional[List[StageStatus]] = None,
        details: bool = False,
        with_summaries: bool = False,
        quantiles: str = "0.05, 0.25, 0.5, 0.75, 0.95",
        task_status: Optional[List[TaskStatus]] = None,
    ) -> List[StageData]:
        """
        Get a list of all stages for an application.

        Args:
            app_id: The application ID
            status: Filter by stage status
            details: Whether to include task details (WARNING: Setting this to True can significantly slow down the API call due to the large amount of task data returned)
            with_summaries: Whether to include summary metrics
            quantiles: Comma-separated list of quantiles to use for summary metrics
            task_status: Filter by task status (only takes effect when details=true)

        Returns:
            List of StageData objects
        """
        params = {
            "details": str(details).lower(),
            "withSummaries": str(with_summaries).lower(),
            "quantiles": quantiles,
        }

        if status:
            params["status"] = [s.value for s in status]
        # taskStatus parameter only takes effect when details=true
        if task_status and details:
            params["taskStatus"] = [s.value for s in task_status]

        data = self._get(f"applications/{app_id}/stages", params)
        return self._parse_model_list(data, StageData)

    def list_stage_attempts(
        self,
        app_id: str,
        stage_id: int,
        details: bool = False,  # Setting this to true is NOT recommended due to the amount of data returned.
        task_status: Optional[List[TaskStatus]] = None,
        with_summaries: bool = True,
        quantiles: str = "0.05, 0.25, 0.5, 0.75, 0.95",
    ) -> List[StageData]:
        """
        Get information about a specific stage.

        Args:
            app_id: The application ID
            stage_id: The stage ID
            details: Whether to include task details
            task_status: Filter by task status
            with_summaries: Whether to include summary metrics
            quantiles: Comma-separated list of quantiles to use for summary metrics

        Returns:
            List of StageData objects (one per attempt)
        """
        params = {
            "details": str(details).lower(),
            "withSummaries": str(with_summaries).lower(),
            "quantiles": quantiles,
        }

        if task_status:
            params["taskStatus"] = [s.value for s in task_status]

        data = self._get(f"applications/{app_id}/stages/{stage_id}", params)
        return self._parse_model_list(data, StageData)

    def get_stage_attempt(
        self,
        app_id: str,
        stage_id: int,
        attempt_id: int,
        details: bool = True,
        task_status: Optional[List[TaskStatus]] = None,
        with_summaries: bool = False,
        quantiles: str = "0.05, 0.25, 0.5, 0.75, 0.95",
    ) -> StageData:
        """
        Get information about a specific stage attempt.

        Args:
            app_id: The application ID
            stage_id: The stage ID
            attempt_id: The attempt ID
            details: Whether to include task details
            task_status: Filter by task status
            with_summaries: Whether to include summary metrics
            quantiles: Comma-separated list of quantiles to use for summary metrics

        Returns:
            StageData object
        """
        params = {
            "details": str(details).lower(),
            "withSummaries": str(with_summaries).lower(),
            "quantiles": quantiles,
        }

        if task_status:
            params["taskStatus"] = [s.value for s in task_status]

        data = self._get(
            f"applications/{app_id}/stages/{stage_id}/{attempt_id}", params
        )
        return self._parse_model(data, StageData)

    def get_stage_task_summary(
        self,
        app_id: str,
        stage_id: int,
        attempt_id: int,
        quantiles: str = "0.05, 0.25, 0.5, 0.75, 0.95",
    ) -> TaskMetricDistributions:
        """
        Get task summary metrics for a specific stage attempt.

        Args:
            app_id: The application ID
            stage_id: The stage ID
            attempt_id: The attempt ID
            quantiles: Comma-separated list of quantiles to use for summary metrics

        Returns:
            TaskMetricDistributions object
        """
        params = {"quantiles": quantiles}
        data = self._get(
            f"applications/{app_id}/stages/{stage_id}/{attempt_id}/taskSummary", params
        )
        return self._parse_model(data, TaskMetricDistributions)

    def list_stage_tasks(
        self,
        app_id: str,
        stage_id: int,
        attempt_id: int,
        offset: int = 0,
        length: int = 20,
        sort_by: str = "ID",
        status: Optional[List[TaskStatus]] = None,
    ) -> List[TaskData]:
        """
        Get tasks for a specific stage attempt.

        Args:
            app_id: The application ID
            stage_id: The stage ID
            attempt_id: The attempt ID
            offset: Pagination offset
            length: Number of tasks to return
            sort_by: Field to sort by
            status: Filter by task status

        Returns:
            List of TaskData objects
        """
        params = {"offset": offset, "length": length, "sortBy": sort_by}

        if status:
            params["status"] = [s.value for s in status]

        data = self._get(
            f"applications/{app_id}/stages/{stage_id}/{attempt_id}/taskList", params
        )
        return self._parse_model_list(data, TaskData)

    def list_executors(self, app_id: str) -> List[ExecutorSummary]:
        """
        Get a list of all executors for an application.

        Args:
            app_id: The application ID

        Returns:
            List of ExecutorSummary objects
        """
        data = self._get(f"applications/{app_id}/executors")
        return self._parse_model_list(data, ExecutorSummary)

    def list_all_executors(self, app_id: str) -> List[ExecutorSummary]:
        """
        Get a list of all executors (active and inactive) for an application.

        Args:
            app_id: The application ID

        Returns:
            List of ExecutorSummary objects
        """
        data = self._get(f"applications/{app_id}/allexecutors")
        return self._parse_model_list(data, ExecutorSummary)

    def list_executor_thread_dump(
        self, app_id: str, executor_id: str
    ) -> List[ThreadStackTrace]:
        """
        Get thread dump for a specific executor.

        Args:
            app_id: The application ID
            executor_id: The executor ID

        Returns:
            List of ThreadStackTrace objects
        """
        data = self._get(f"applications/{app_id}/executors/{executor_id}/threads")
        return self._parse_model_list(data, ThreadStackTrace)

    def get_task_thread_dump(
        self, app_id: str, task_id: int, executor_id: str
    ) -> ThreadStackTrace:
        """
        Get thread dump for a specific task.

        Args:
            app_id: The application ID
            task_id: The task ID
            executor_id: The executor ID

        Returns:
            ThreadStackTrace object
        """
        params = {"taskId": task_id, "executorId": executor_id}
        data = self._get(f"applications/{app_id}/threads", params)
        return self._parse_model(data, ThreadStackTrace)

    def list_all_processes(self, app_id: str) -> List[ProcessSummary]:
        """
        Get a list of all processes for an application.

        Args:
            app_id: The application ID

        Returns:
            List of ProcessSummary objects
        """
        data = self._get(f"applications/{app_id}/allmiscellaneousprocess")
        return self._parse_model_list(data, ProcessSummary)

    def list_rdds(self, app_id: str) -> List[RDDStorageInfo]:
        """
        Get a list of all RDDs for an application.

        Args:
            app_id: The application ID

        Returns:
            List of RDDStorageInfo objects
        """
        data = self._get(f"applications/{app_id}/storage/rdd")
        return self._parse_model_list(data, RDDStorageInfo)

    def get_rdd(self, app_id: str, rdd_id: int) -> RDDStorageInfo:
        """
        Get information about a specific RDD.

        Args:
            app_id: The application ID
            rdd_id: The RDD ID

        Returns:
            RDDStorageInfo object
        """
        data = self._get(f"applications/{app_id}/storage/rdd/{rdd_id}")
        return self._parse_model(data, RDDStorageInfo)

    def get_environment(self, app_id: str) -> ApplicationEnvironmentInfo:
        """
        Get environment information for an application.

        Args:
            app_id: The application ID

        Returns:
            ApplicationEnvironmentInfo object
        """
        data = self._get(f"applications/{app_id}/environment")
        return self._parse_model(data, ApplicationEnvironmentInfo)

    def get_metrics_prometheus(self, app_id: str) -> str:
        """
        Get Prometheus metrics for an application.

        Args:
            app_id: The application ID

        Returns:
            Prometheus metrics as a string
        """
        url = urljoin(
            self.base_url.replace("/api/v1", "/metrics/executors"), "prometheus"
        )

        if self.session:
            response = self.session.get(url, timeout=self.timeout, proxies=self.proxies)
        else:
            response = requests.get(url, timeout=self.timeout, proxies=self.proxies)

        response.raise_for_status()
        return response.text

    def get_sql_list(
        self,
        app_id: str,
        attempt_id: Optional[str] = None,
        details: bool = True,
        plan_description: bool = False,
        offset: int = 0,
        length: int = 20,
    ) -> List[ExecutionData]:
        """
        Get a list of all SQL executions for an application.

        Args:
            app_id: The application ID
            attempt_id: Optional attempt ID
            details: Whether to include execution details
            plan_description: Whether to include plan description
            offset: Pagination offset
            length: Number of executions to return

        Returns:
            List of ExecutionData objects
        """
        params = {
            "details": str(details).lower(),
            "planDescription": str(plan_description).lower(),
            "offset": offset,
            "length": length,
        }

        if attempt_id:
            endpoint = f"applications/{app_id}/{attempt_id}/sql"
        else:
            endpoint = f"applications/{app_id}/sql"

        data = self._get(endpoint, params)
        return [ExecutionData.from_dict(item) for item in data]

    def get_sql_execution(
        self,
        app_id: str,
        execution_id: int,
        attempt_id: Optional[str] = None,
        details: bool = True,
        plan_description: bool = True,
    ) -> ExecutionData:
        """
        Get information about a specific SQL execution.

        Args:
            app_id: The application ID
            execution_id: The execution ID
            attempt_id: Optional attempt ID
            details: Whether to include execution details
            plan_description: Whether to include plan description

        Returns:
            ExecutionData object
        """
        params = {
            "details": str(details).lower(),
            "planDescription": str(plan_description).lower(),
        }

        if attempt_id:
            endpoint = f"applications/{app_id}/{attempt_id}/sql/{execution_id}"
        else:
            endpoint = f"applications/{app_id}/sql/{execution_id}"

        data = self._get(endpoint, params)
        return ExecutionData.from_dict(data)



#!/usr/bin/env python3
"""
EMR Persistent App UI Client

This module provides functionality to create an EMR Persistent App UI, retrieve its details and presigned URL,
and establish an HTTP session with proper cookie management for Spark History Server access.
"""

import logging
import time
from typing import Dict, Optional, Tuple
from urllib.parse import urlparse

import boto3
import requests
from botocore.exceptions import ClientError

from spark_history_mcp.config.config import ServerConfig

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


class EMRPersistentUIClient:
    """Client for managing EMR Persistent App UI and HTTP sessions."""

    def __init__(self, server_config: ServerConfig):
        """
        Initialize the EMR client.

        Args:
            server_config: ServerConfig object
        """
        self.emr_cluster_arn = server_config.emr_cluster_arn
        self.region = self.emr_cluster_arn.split(":")[3]  # Extract region from ARN

        # Initialize boto3 client with credentials
        self.emr_client = boto3.client(
            "emr",
            region_name=self.region,
        )

        self.session = requests.Session()
        self.persistent_ui_id: Optional[str] = None
        self.presigned_url: Optional[str] = None
        self.base_url: Optional[str] = None
        self.timeout: int = server_config.timeout

    def create_persistent_app_ui(self) -> Dict:
        """
        Create a persistent app UI for the given cluster.

        Returns:
            Response from create-persistent-app-ui API call

        Raises:
            ClientError: If the API call fails
        """
        logger.info(f"Creating persistent app UI for cluster: {self.emr_cluster_arn}")

        try:
            response = self.emr_client.create_persistent_app_ui(
                TargetResourceArn=self.emr_cluster_arn
            )

            self.persistent_ui_id = response.get("PersistentAppUIId")
            runtime_role_enabled = response.get("RuntimeRoleEnabledCluster", False)

            logger.info("✅ Persistent App UI created successfully")
            logger.info(f"   Persistent UI ID: {self.persistent_ui_id}")
            logger.info(f"   Runtime Role Enabled: {runtime_role_enabled}")

            return response

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            error_message = e.response["Error"]["Message"]
            logger.error(
                f"❌ Failed to create persistent app UI: {error_code} - {error_message}"
            )
            raise
        except Exception as e:
            logger.error(f"❌ Unexpected error creating persistent app UI: {str(e)}")
            raise

    def describe_persistent_app_ui(self) -> Dict:
        """
        Describe the persistent app UI.

        Returns:
            Response from describe-persistent-app-ui API call

        Raises:
            ValueError: If no persistent UI ID is available
            ClientError: If the API call fails
        """
        if not self.persistent_ui_id:
            raise ValueError("No persistent UI ID available. Create one first.")

        logger.info(f"Describing persistent app UI: {self.persistent_ui_id}")

        try:
            response = self.emr_client.describe_persistent_app_ui(
                PersistentAppUIId=self.persistent_ui_id
            )

            ui_details = response.get("PersistentAppUI", {})
            logger.info("✅ Persistent App UI details retrieved")
            logger.info(f"   Status: {ui_details.get('PersistentAppUIStatus')}")
            logger.info(f"   Creation Time: {ui_details.get('CreationTime')}")

            return response

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            error_message = e.response["Error"]["Message"]
            logger.error(
                f"❌ Failed to describe persistent app UI: {error_code} - {error_message}"
            )
            raise
        except Exception as e:
            logger.error(f"❌ Unexpected error describing persistent app UI: {str(e)}")
            raise

    def get_presigned_url(self, ui_type: str = "SHS") -> str:
        """
        Get presigned URL for the persistent app UI.

        Args:
            ui_type: Type of UI ('SHS' for Spark History Server)

        Returns:
            Presigned URL string

        Raises:
            ValueError: If no persistent UI ID is available
            ClientError: If the API call fails
        """
        if not self.persistent_ui_id:
            raise ValueError("No persistent UI ID available. Create one first.")

        logger.info(
            f"Getting presigned URL for persistent app UI: {self.persistent_ui_id} (type: {ui_type})"
        )

        try:
            response = self.emr_client.get_persistent_app_ui_presigned_url(
                PersistentAppUIId=self.persistent_ui_id, PersistentAppUIType=ui_type
            )

            self.presigned_url_ready = response.get("PresignedURLReady")
            self.presigned_url = response.get("PresignedURL")

            # Extract base URL from presigned URL
            parsed_url = urlparse(self.presigned_url)
            self.base_url = f"{parsed_url.scheme}://{parsed_url.netloc}/shs"

            logger.info("✅ Presigned URL obtained successfully")
            logger.info(f"   Base URL: {self.base_url}")

            return self.presigned_url

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            error_message = e.response["Error"]["Message"]
            logger.error(
                f"❌ Failed to get presigned URL: {error_code} - {error_message}"
            )
            raise
        except Exception as e:
            logger.error(f"❌ Unexpected error getting presigned URL: {str(e)}")
            raise

    def setup_http_session(self) -> requests.Session:
        """
        Set up HTTP session with proper headers and cookie management.

        Returns:
            Configured requests.Session object

        Raises:
            ValueError: If no presigned URL is available
        """
        if not self.presigned_url:
            raise ValueError("No presigned URL available. Get one first.")

        logger.info("Setting up HTTP session with cookie management")

        # Configure session with appropriate headers
        self.session.headers.update(
            {
                "User-Agent": "EMR-Persistent-UI-Client/1.0",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate",
                "Connection": "keep-alive",
                "Upgrade-Insecure-Requests": "1",
            }
        )

        try:
            # Make initial request to establish session and get cookies
            logger.info("Making initial request")
            response = self.session.get(
                self.presigned_url, timeout=self.timeout, allow_redirects=True
            )
            response.raise_for_status()

            logger.info("✅ HTTP session established successfully")
            logger.info(f"   Status Code: {response.status_code}")
            logger.info(f"   Cookies: {len(self.session.cookies)} cookie(s) stored")

            # Log cookie details (without sensitive values)
            for cookie in self.session.cookies:
                logger.debug(f"   Cookie: {cookie.name} (domain: {cookie.domain})")

            return self.session

        except requests.exceptions.RequestException as e:
            logger.error(f"❌ Failed to establish HTTP session: {str(e)}")
            raise
        except Exception as e:
            logger.error(f"❌ Unexpected error setting up HTTP session: {str(e)}")
            raise

    def initialize(self) -> Tuple[str, requests.Session]:
        """
        Initialize the EMR Persistent UI client by creating a persistent app UI,
        verifying its status, getting a presigned URL, and setting up an HTTP session.

        If the status is STARTING, it will wait for the status
        to change to ATTACHED before proceeding.

        Returns:
            Tuple containing the base URL and configured session

        Raises:
            ValueError: If the persistent UI status is not ATTACHED after waiting
        """
        # Step 1: Create persistent app UI
        self.create_persistent_app_ui()

        # Step 2: Describe persistent app UI and verify status
        # Wait for up to 3 minutes if status is STARTING
        max_wait_time = 180  # 3 minutes in seconds
        wait_interval = 10  # Check every 10 seconds
        total_waited = 0
        ui_status = ""

        while total_waited < max_wait_time:
            describe_response = self.describe_persistent_app_ui()
            ui_status = describe_response.get("PersistentAppUI", {}).get(
                "PersistentAppUIStatus"
            )

            if ui_status == "ATTACHED":
                # Status is good, proceed
                break
            elif ui_status == "STARTING":
                # Status is starting, wait and check again
                logger.info(
                    f"EMR Persistent UI status is {ui_status}, waiting for ATTACHED status..."
                )
                time.sleep(wait_interval)
                total_waited += wait_interval
            else:
                # Status is something else (not STARTING or ATTACHED), raise error
                raise ValueError(
                    f"EMR Persistent UI status is {ui_status}, expected ATTACHED or STARTING"
                )

        # After waiting, check if we have the correct status
        if ui_status != "ATTACHED":
            raise ValueError(
                f"EMR Persistent UI status is still {ui_status} after waiting {total_waited} seconds, expected ATTACHED"
            )

        # Step 3: Get presigned URL
        self.get_presigned_url()

        # Step 4: Setup HTTP session
        self.setup_http_session()

        return self.base_url, self.session



import asyncio
import os
import uuid
from urllib.parse import urljoin

from playwright.async_api import async_playwright

from spark_history_mcp.config.config import ServerConfig

# HTML client for Spark History Server web interface


class SparkHtmlClient:
    def __init__(self, server_config: ServerConfig):
        self.config = server_config
        self.base_url = self.config.url.rstrip("/") + "/history/"
        self.auth = None
        self.browser = None

        # Set up authentication if provided
        if self.config.auth:
            if self.config.auth.username and self.config.auth.password:
                self.auth = (self.config.auth.username, self.config.auth.password)

    async def get_rendered_html(self, path):
        """
        Fetches and returns the fully rendered HTML content of a Spark UI page.

        This method uses Playwright to launch a headless browser, navigate to the specified path
        on the Spark UI, wait for the page to fully render (including any JavaScript execution),
        and then return the complete HTML content.

        Args:
            path: The path to navigate to, relative to the base URL

        Returns:
            str: The fully rendered HTML content of the page
        """
        async with async_playwright() as p:
            # Launch browser if not already launched
            if not self.browser:
                self.browser = await p.chromium.launch()
            page = await self.browser.new_page()
            await page.set_viewport_size({"width": 1280, "height": 800})
            url = urljoin(self.base_url, path)
            await page.goto(url)

            # Wait for network to be idle (no more than 2 connections for at least 500 ms)
            await page.wait_for_load_state("networkidle")

            await page.wait_for_timeout(3000)  # 3 seconds

            # Get the fully rendered HTML
            html_content = await page.content()

            return html_content

    async def get_screenshot(self, path, save_path=None):
        """
        Takes a screenshot of the page at the given path and saves it.

        Args:
            path: The path to navigate to for the screenshot
            save_path: The file path where the screenshot should be saved.
                       If None, saves to a random filename in /tmp.

        Returns:
            The full path to the saved screenshot
        """
        async with async_playwright() as p:
            if not self.browser:
                self.browser = await p.chromium.launch()
            path = path.lstrip("/")
            page = await self.browser.new_page()
            await page.set_viewport_size({"width": 2560, "height": 800})
            url = urljoin(self.base_url, path)
            await page.goto(url)

            # Wait for network to be idle
            await page.wait_for_load_state("networkidle")
            await page.wait_for_timeout(3000)  # 3 seconds

            # Use provided save_path or generate a random filename
            import tempfile

            filename = (
                save_path
                if save_path
                else f"{tempfile.gettempdir()}/{uuid.uuid4()}.jpg"
            )

            # Ensure directory exists
            os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
            await page.screenshot(
                path=filename, type="jpeg", quality=100, full_page=True
            )
            return filename


async def main():
    client = SparkHtmlClient(ServerConfig(url="http://localhost:18080"))

    # Example with custom save path
    await client.get_screenshot(
        "/spark-e975c1c221934381b99772c54ae4b8e6/jobs/", "spark_executors.jpg"
    )
    # html = await client.get_rendered_html("spark-e975c1c221934381b99772c54ae4b8e6/executors/")

    # Close the browser when done with all operations
    if client.browser:
        await client.browser.close()

    # Save the HTML to a file
    # with open("rendered_page.html", "w", encoding="utf-8") as f:
    #     f.write(html)
    #
    # print("HTML content saved to rendered_page.html")


if __name__ == "__main__":
    asyncio.run(main())



import os
from typing import Any, Dict, List, Literal, Optional, Tuple

import yaml
from pydantic import Field
from pydantic.fields import FieldInfo
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)


class YamlConfigSettingsSource(PydanticBaseSettingsSource):
    """Custom settings source that loads configuration from a YAML file.

    The file path is determined by the SHS_MCP_CONFIG environment variable,
    defaulting to 'config.yaml' if not set.
    """

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> Tuple[Any, str, bool]:
        # Not used for this implementation
        return None, field_name, False

    def __call__(self) -> Dict[str, Any]:
        """Load and return the YAML configuration data."""
        config_path = os.getenv("SHS_MCP_CONFIG", "config.yaml")
        is_explicitly_set = "SHS_MCP_CONFIG" in os.environ

        if not os.path.exists(config_path):
            # If the config file was explicitly specified but doesn't exist, fail fast
            if is_explicitly_set:
                raise FileNotFoundError(
                    f"Config file not found: {config_path}\n"
                    f"Specified via: SHS_MCP_CONFIG environment variable"
                )
            # If using default and it doesn't exist, return empty (will use defaults)
            return {}

        with open(config_path, "r") as f:
            config_data = yaml.safe_load(f)

        return config_data or {}


class AuthConfig(BaseSettings):
    """Authentication configuration for the Spark server."""

    username: Optional[str] = Field(None)
    password: Optional[str] = Field(None)
    token: Optional[str] = Field(None)


class ServerConfig(BaseSettings):
    """Server configuration for the Spark server."""

    url: Optional[str] = None
    auth: AuthConfig = Field(default_factory=AuthConfig, exclude=True)
    default: bool = False
    verify_ssl: bool = True
    emr_cluster_arn: Optional[str] = None  # EMR specific field
    use_proxy: bool = False
    timeout: int = 30  # HTTP request timeout in seconds
    include_plan_description: Optional[bool] = None


class McpConfig(BaseSettings):
    """Configuration for the MCP server."""

    transports: List[Literal["stdio", "sse", "streamable-http"]] = Field(
        default_factory=list
    )
    address: Optional[str] = "localhost"
    port: Optional[int | str] = "18888"
    debug: Optional[bool] = False
    model_config = SettingsConfigDict(extra="ignore")


class Config(BaseSettings):
    """Configuration for the Spark client."""

    servers: Dict[str, ServerConfig] = {
        "local": ServerConfig(url="http://localhost:18080", default=True),
    }
    mcp: Optional[McpConfig] = McpConfig(transports=["streamable-http"])
    model_config = SettingsConfigDict(
        env_prefix="SHS_",
        env_nested_delimiter="_",
        env_file=".env",
        env_file_encoding="utf-8",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # Precedence order (highest to lowest):
        # 1. Environment variables
        # 2. .env file
        # 3. YAML config file (from SHS_MCP_CONFIG)
        # 4. Init settings (constructor arguments)
        # 5. File secrets
        return (
            env_settings,
            dotenv_settings,
            YamlConfigSettingsSource(settings_cls),
            init_settings,
            file_secret_settings,
        )



"""Main entry point for Spark History Server MCP."""

import argparse
import json
import logging
import os
import sys

from spark_history_mcp.config.config import Config
from spark_history_mcp.core import app

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


def main():
    """Main entry point."""
    parser = argparse.ArgumentParser(description="Spark History Server MCP")
    parser.add_argument(
        "--config",
        "-c",
        default=os.getenv("SHS_MCP_CONFIG", "config.yaml"),
        help="Path to config file (default: config.yaml, env: SHS_MCP_CONFIG)",
    )
    args = parser.parse_args()

    try:
        logger.info("Starting Spark History Server MCP...")
        logger.info(f"Using config file: {args.config}")

        # Set the config file path in environment for Pydantic Settings
        os.environ["SHS_MCP_CONFIG"] = args.config

        # Now Config() will automatically load from the specified YAML file
        config = Config()
        if config.mcp.debug:
            logger.setLevel(logging.DEBUG)
        logger.debug(json.dumps(json.loads(config.model_dump_json()), indent=4))
        app.run(config)
    except FileNotFoundError as e:
        logger.error(f"Configuration error: {e}")
        sys.exit(1)
    except Exception as e:
        logger.error(f"Failed to start MCP server: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()
